from functools import lru_cache
from pathlib import Path

import mlx.core as mx

_SOURCE = (Path(__file__).parent / "rotation.metal").read_text()
_MAX_GROUP_SIZE = 128
_MAX_KROT = 16


@lru_cache(maxsize=None)
def get_rotation_kernel(rows_per_tile: int):
    """Compile and cache the Metal rotation kernel for a given tile size."""
    return mx.fast.metal_kernel(
        name=f"paro_rotate_r{rows_per_tile}",
        input_names=["x", "packed_pairs", "cos_theta", "sin_theta", "channel_scales", "params"],
        output_names=["out"],
        source=_SOURCE.format(
            ROWS_PER_TILE=rows_per_tile,
            MAX_GROUP_SIZE=_MAX_GROUP_SIZE,
            MAX_KROT=_MAX_KROT,
        ),
    )



from __future__ import annotations

import itertools
from collections.abc import AsyncIterator
from dataclasses import asdict

import torch.distributed as dist
from transformers import AutoTokenizer
from vllm import AsyncEngineArgs, AsyncLLMEngine, SamplingParams

import paroquant.inference.backends.vllm.plugin  # noqa: F401 — registers quantization config
from paroquant.inference.base import BaseGenerator, GenerationParams


class VllmGenerator(BaseGenerator):
    backend = "vllm"

    def __init__(self, model: str, **engine_kwargs):
        self.engine = AsyncLLMEngine.from_engine_args(AsyncEngineArgs(model=model, **engine_kwargs))
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self._req_counter = itertools.count()

    async def stream_generate(self, prompt: str, params: GenerationParams) -> AsyncIterator[str]:
        sampling = SamplingParams(**asdict(params))
        last_len = 0

        async for output in self.engine.generate(prompt, sampling, f"req-{next(self._req_counter)}"):
            text = output.outputs[0].text
            delta = text[last_len:]
            last_len = len(text)
            if delta:
                yield delta

    async def close(self) -> None:
        shutdown = getattr(self.engine, "shutdown", None) or getattr(self.engine, "shutdown_background_loop", None)
        if shutdown:
            shutdown()

        if dist.is_initialized():
            dist.destroy_process_group()



from __future__ import annotations

import random
from collections.abc import Callable, Iterator
from copy import deepcopy
from typing import Literal

import torch
import torch.nn as nn
from tqdm import tqdm

from .util import CosineAnnealingParam, logger


def _get_independent_channel_pairs(
    pairs: torch.Tensor, dim: int, num_rotations: int, num_pairs_each: int
) -> list[list[tuple[int, int]]]:
    pairs = pairs.cpu().tolist()
    rotations_pairs = []
    # Record the available pairs in a matrix.
    available = torch.ones(dim, dim)
    available.fill_diagonal_(0)

    for _ in range(num_rotations):
        independent_pairs = []
        # We use a copy to track the available pairs in this rotation.
        # It's different from the original available matrix.
        available_in_rotation = available.clone()
        # Greedily select pairs that are independent
        for i, j in pairs:
            if len(independent_pairs) == num_pairs_each:
                break
            if available_in_rotation[i, j] == 0:
                continue
            # Simply select the first available pair.
            independent_pairs.append((i, j))
            # Selecting (i, j) in this rotation prohibits
            # selecting any other pairs that share i or j
            # in this rotation.
            available_in_rotation[i, :] = 0
            available_in_rotation[j, :] = 0
            available_in_rotation[:, i] = 0
            available_in_rotation[:, j] = 0
            # Mark the pair as unavailable for future selections.
            # i and j are still available for any other pairs
            # in next rotations.
            available[i, j] = 0
            available[j, i] = 0

        rotations_pairs.append(independent_pairs)

    return rotations_pairs


def get_random_rotation_pairs(
    sensitivity_input: torch.Tensor,
    group_size: int,
    num_rotations: int,
    num_pairs_factor: float,
    seed: int,
) -> list[tuple[int, int]]:
    sorted_pairs: list[list[tuple[int, int]]] = []
    rand = random.Random(seed)
    group_num = sensitivity_input.shape[0]

    for group_idx in range(group_num):
        sorted_pairs.append([])
        for i in range(group_size):
            for j in range(i + 1, group_size):
                sorted_pairs[group_idx].append((i, j))

        rand.shuffle(sorted_pairs[group_idx])

    sorted_pairs = torch.tensor(sorted_pairs, device=sensitivity_input.device)

    pairs_k_groups = [[] for _ in range(num_rotations)]
    num_pairs_per_group = int(group_size * num_pairs_factor)

    for i in range(group_num):
        offset = i * group_size
        pairs_g_k_groups = _get_independent_channel_pairs(
            sorted_pairs[i], group_size, num_rotations, num_pairs_per_group
        )
        for r_idx in range(num_rotations):
            pairs_g = pairs_g_k_groups[r_idx]
            for j, (col1, col2) in enumerate(pairs_g):
                pairs_g[j] = (col1 + offset, col2 + offset)
            pairs_k_groups[r_idx].extend(pairs_g)

    return pairs_k_groups


def optimize_module(
    module: nn.Module,
    train_set_batches: tuple[Iterator[torch.Tensor], Iterator[torch.Tensor]],
    val_set_batches: tuple[Iterator[torch.Tensor], Iterator[torch.Tensor]],
    kwargs: dict,
    optim_params: list[dict],
    *,
    loss_fn: Literal["mse", "smooth_l1"],
    n_iter: int,
    early_stop: int | None,
    post_optim_callback: Callable[[nn.Module], None] | None = None,
) -> None:
    train_input_batches, train_output_batches = train_set_batches
    val_input_batches, val_output_batches = val_set_batches

    total_steps = n_iter * len(train_input_batches)
    schedulers = [
        CosineAnnealingParam(
            start_value=param_group["lr"],
            end_value=param_group["lr"] / 20,
            T_max=total_steps,
        )
        for param_group in optim_params
    ]
    optimizer = torch.optim.AdamW(optim_params)
    scaler = torch.amp.GradScaler()
    if loss_fn == "mse":
        loss = nn.MSELoss()
    elif loss_fn == "smooth_l1":
        loss = nn.SmoothL1Loss()
    else:
        raise ValueError(f"Unsupported loss function: {loss_fn}")

    progress_bar = tqdm(total=n_iter, unit="iter")

    def module_output(input: torch.Tensor) -> torch.Tensor:
        out = module(input, **kwargs)
        if isinstance(out, tuple):
            out = out[0]
        return out

    @torch.no_grad()
    def loss_batches(input_batches: Iterator[torch.Tensor], output_batches: Iterator[torch.Tensor]) -> torch.Tensor:
        total_loss = None
        for input_batch, output_batch in zip(input_batches, output_batches):
            output_q = module_output(input_batch)
            loss_value = loss(output_batch, output_q)
            if total_loss is None:
                total_loss = loss_value
            else:
                total_loss += loss_value
        return total_loss / len(input_batches)

    with torch.no_grad():
        original_val_loss = loss_batches(val_input_batches, val_output_batches)

    best_val_loss = original_val_loss
    best_sd = deepcopy(module.state_dict())

    early_stop_counter = 0

    for _ in range(n_iter):
        for input_batch, output_batch in zip(train_input_batches, train_output_batches):
            optimizer.zero_grad()
            with torch.amp.autocast("cuda"):
                output_q = module_output(input_batch)
                loss_value = loss(output_batch, output_q)

            scaler.scale(loss_value).backward()
            scaler.step(optimizer)
            scaler.update()

            for i, scheduler in enumerate(schedulers):
                optimizer.param_groups[i]["lr"] = scheduler.step()

            if post_optim_callback:
                post_optim_callback(module)

        with torch.no_grad():
            val_loss_value = loss_batches(val_input_batches, val_output_batches)

        if val_loss_value < best_val_loss:
            best_val_loss = val_loss_value
            best_sd = deepcopy(module.state_dict())
            early_stop_counter = 0
        else:
            early_stop_counter += 1
            if early_stop is not None and early_stop_counter >= early_stop:
                break

        progress_bar.set_postfix(
            val_loss=val_loss_value.item(),
            val_og_loss=original_val_loss.item(),
        )
        progress_bar.update(1)

    progress_bar.close()
    logger.info(f"Best val loss: {best_val_loss.item()}, Original val loss: {original_val_loss.item()}")

    module.load_state_dict(best_sd)



import math

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from paroquant.kernels.metal import get_rotation_kernel


def _pack_pairs(pairs: mx.array, group_size: int) -> mx.array:
    """Pack int16 pair indices into int32 for the Metal kernel."""
    krot, hidden = int(pairs.shape[0]), int(pairs.shape[1])
    p = np.array(pairs, copy=False).reshape(krot, hidden // group_size, group_size).astype(np.int32, copy=False)
    return mx.array((p[:, :, 0::2] | (p[:, :, 1::2] << 16)).reshape(krot, -1))


class RotateQuantizedLinear(nn.Module):
    """Pairwise Givens rotation + quantized matmul (Metal kernel)."""

    def __init__(
        self,
        input_dims: int,
        output_dims: int,
        bias: bool = True,
        group_size: int = 128,
        bits: int = 4,
        krot: int = 8,
    ):
        super().__init__()
        self.group_size = group_size
        self.bits = bits

        self.theta = mx.zeros((krot, input_dims // 2))
        self.pairs = mx.zeros((krot, input_dims), dtype=mx.int16)
        self.channel_scales = mx.ones((1, input_dims))

        self.weight = mx.zeros((output_dims, input_dims * bits // 32), dtype=mx.uint32)
        self.scales = mx.zeros((output_dims, input_dims // group_size))
        self.biases = mx.zeros((output_dims, input_dims // group_size))

        if bias:
            self.bias = mx.zeros((output_dims,))

        self._cached = False

    def _cache_rotation(self):
        """Pre-compute sin/cos and pack pairs (called once on first forward)."""
        dim = self.theta.shape[1] * 2
        self._dim = dim
        self._half_group = self.group_size // 2
        self._num_groups = dim // self.group_size
        self._krot = int(self.theta.shape[0])
        self._cos = mx.cos(self.theta)
        self._sin = mx.sin(self.theta)
        self._packed_pairs = _pack_pairs(self.pairs, self.group_size)
        self._scales_flat = self.channel_scales.reshape(-1)
        self._cached = True

    def _rotate(self, x: mx.array) -> mx.array:
        batch = x.shape[0]
        tile = 1 if batch <= 1 else 4
        params = mx.array([batch, self._dim, self._krot, self.group_size], dtype=mx.int32)
        grid = (math.ceil(batch / tile) * self._half_group, self._num_groups, 1)
        return get_rotation_kernel(tile)(
            inputs=[x, self._packed_pairs, self._cos, self._sin, self._scales_flat, params],
            output_shapes=[x.shape],
            output_dtypes=[x.dtype],
            grid=grid,
            threadgroup=(self._half_group, 1, 1),
        )[0]

    def __call__(self, x: mx.array) -> mx.array:
        if not self._cached:
            self._cache_rotation()

        shape = x.shape
        rotated = self._rotate(x.reshape(-1, self._dim))

        y = mx.quantized_matmul(
            rotated.reshape(shape),
            self.weight,
            scales=self.scales,
            biases=self.biases,
            transpose=True,
            group_size=self.group_size,
            bits=self.bits,
        )
        if "bias" in self:
            y = y + self.bias
        return y



from __future__ import annotations

import glob
import json
import logging
import os
from typing import TYPE_CHECKING

import torch
import torch.nn as nn
from huggingface_hub import snapshot_download
from safetensors import safe_open
from transformers.quantizers.auto import (
    register_quantization_config,
    register_quantizer,
)
from transformers.quantizers.base import HfQuantizer
from transformers.utils.quantization_config import QuantizationConfigMixin

import paroquant.kernels.cuda  # noqa: F401 — registers torch.ops.rotation.rotate

from .modules import RotateQuantizedLinear

if TYPE_CHECKING:
    from transformers import PreTrainedModel

logger = logging.getLogger(__name__)


def _find_quantized_modules(model_path: str) -> set[str]:
    """Scan checkpoint safetensors to find modules that have ``.qweight`` keys."""
    local_dir = model_path if os.path.isdir(model_path) else snapshot_download(model_path)

    index_file = os.path.join(local_dir, "model.safetensors.index.json")
    if os.path.exists(index_file):
        with open(index_file) as f:
            keys = json.load(f).get("weight_map", {}).keys()
    else:
        keys = []
        for sf in sorted(glob.glob(os.path.join(local_dir, "*.safetensors"))):
            with safe_open(sf, framework="pt") as st:
                keys.extend(st.keys())

    return {k.rsplit(".", 1)[0] for k in keys if k.endswith(".qweight")}


@register_quantization_config("paroquant")
class ParoQuantConfig(QuantizationConfigMixin):
    """Quantization config for ParoQuant checkpoints."""

    def __init__(
        self,
        bits: int = 4,
        group_size: int = 128,
        krot: int = 8,
        modules_to_not_convert: list[str] | None = None,
        **kwargs,
    ):
        self.quant_method = "paroquant"
        self.bits = bits
        self.group_size = group_size
        self.krot = krot
        self.modules_to_not_convert = modules_to_not_convert
        if hasattr(self, "post_init"):
            self.post_init()


@register_quantizer("paroquant")
class ParoQuantHfQuantizer(HfQuantizer):
    """Replaces nn.Linear with RotateQuantizedLinear for quantized layers.

    Only modules with ``.qweight`` in the checkpoint are replaced — visual
    encoders and other unquantized layers are left untouched automatically.
    """

    requires_calibration = True

    def validate_environment(self, **kwargs):
        if not torch.cuda.is_available():
            raise RuntimeError("ParoQuant requires CUDA.")

    def update_dtype(self, dtype):
        if dtype != torch.float16:
            logger.warning("ParoQuant requires float16. Overriding dtype=%s → float16.", dtype)
            return torch.float16
        return dtype

    def _process_model_before_weight_loading(self, model: "PreTrainedModel", **kwargs):
        qcfg = self.quantization_config
        quantized_modules = _find_quantized_modules(model.config._name_or_path)
        if qcfg.modules_to_not_convert:
            quantized_modules -= set(qcfg.modules_to_not_convert)
        logger.info("Found %d quantized modules in checkpoint.", len(quantized_modules))

        for name, module in model.named_modules():
            if not isinstance(module, nn.Linear):
                continue
            if name not in quantized_modules:
                continue

            parent_name, attr = name.rsplit(".", 1) if "." in name else ("", name)
            parent = model.get_submodule(parent_name) if parent_name else model

            setattr(
                parent,
                attr,
                RotateQuantizedLinear(
                    module.in_features,
                    module.out_features,
                    bias=module.bias is not None,
                    group_size=qcfg.group_size,
                    bits=qcfg.bits,
                    krot=qcfg.krot,
                ),
            )

    @property
    def is_trainable(self) -> bool:
        return False

    def is_serializable(self) -> bool:
        return True



from pathlib import Path

import torch
from torch.utils.cpp_extension import load

_dir = Path(__file__).parent
_C = load(
    name="paroquant_rotation",
    sources=[str(_dir / "pybind.cpp"), str(_dir / "rotation.cu")],
    extra_cuda_cflags=[
        "-O3",
        "-std=c++17",
        "-U__CUDA_NO_HALF_OPERATORS__",
        "-U__CUDA_NO_HALF_CONVERSIONS__",
        "-U__CUDA_NO_BFLOAT16_OPERATORS__",
        "-U__CUDA_NO_BFLOAT16_CONVERSIONS__",
        "--expt-relaxed-constexpr",
        "--expt-extended-lambda",
        "--use_fast_math",
    ],
    extra_cflags=["-O2", "-std=c++17"],
    verbose=False,
)

try:

    @torch.library.register_fake("rotation::rotate")
    def _fake_kernel(x, idx_ij, theta, scales=None, group_size=128):
        return torch.empty_like(x)

except RuntimeError:
    pass

from .autograd import RotateTensorFunc, scaled_pairwise_rotation  # noqa: E402

__all__ = ["scaled_pairwise_rotation", "RotateTensorFunc"]



from __future__ import annotations

import json
from argparse import ArgumentParser
from pathlib import Path
from typing import Any

import torch
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoModelForImageTextToText, AutoTokenizer

from paroquant.optim.util import get_named_linears, set_module_by_name

_AWQ_REORDER = (0, 2, 4, 6, 1, 3, 5, 7)
_LAYER_PATHS = ["model.layers", "model.language_model.layers", "language_model.layers"]


def _load_model(model_id: str, device_map: str = "cpu") -> torch.nn.Module:
    kwargs = dict(torch_dtype=torch.float16, device_map=device_map, low_cpu_mem_usage=True, trust_remote_code=True)
    try:
        return AutoModelForImageTextToText.from_pretrained(model_id, **kwargs)
    except (ValueError, KeyError):
        return AutoModelForCausalLM.from_pretrained(model_id, **kwargs)


def _get_blocks(model: torch.nn.Module):
    for path in _LAYER_PATHS:
        node = model
        try:
            for attr in path.split("."):
                node = getattr(node, attr)
            return node
        except AttributeError:
            continue
    raise NotImplementedError(f"Unsupported model structure: {type(model)}")


def _get_value(state_dict: dict, *keys: str):
    for key in keys:
        if key in state_dict:
            val = state_dict[key]
            return int(val.item()) if isinstance(val, torch.Tensor) else val
    raise KeyError(f"None of {keys} found")


def _stack_if_numbered(state_dict: dict, key: str) -> torch.Tensor:
    if key in state_dict:
        return state_dict[key]
    parts = []
    i = 0
    while f"{key}.{i}" in state_dict:
        parts.append(state_dict[f"{key}.{i}"])
        i += 1
    if parts:
        return torch.stack(parts)
    raise KeyError(key)


def _pack_awq(values: torch.Tensor, bits: int = 4) -> torch.Tensor:
    pack_factor = 32 // bits
    reordered = values.to(torch.int32).view(values.shape[0], -1, pack_factor)[:, :, _AWQ_REORDER]
    packed = torch.zeros(reordered.shape[0], reordered.shape[1], dtype=torch.int32, device=values.device)
    for i in range(pack_factor):
        packed |= (reordered[:, :, i] & 0xF) << (bits * i)
    return packed


@torch.no_grad()
def _convert_pseudo(model: torch.nn.Module, result_dir: Path) -> int:
    from paroquant.optim.qlinear import PseudoQuantizedLinear

    blocks = _get_blocks(model)
    count = 0
    for layer_idx, layer in enumerate(tqdm(blocks, desc="Pseudo-quantizing")):
        layer = layer.cuda()
        for name, module in get_named_linears(layer).items():
            pt_file = result_dir / f"{layer_idx}.{name}.pt"
            if not pt_file.exists():
                continue
            sd = torch.load(pt_file, weights_only=False, map_location="cuda")
            module.weight.data.copy_(PseudoQuantizedLinear.from_state_dict(sd).pseudo_weight())
            count += 1
        layer.cpu()
    return count


@torch.no_grad()
def _quantize_layer(state_dict: dict, device: str) -> tuple[dict[str, torch.Tensor], int, int, int]:
    from paroquant.kernels.cuda import scaled_pairwise_rotation

    weight = state_dict["weight"].to(device=device, dtype=torch.float32)
    out_features, in_features = weight.shape

    bits = int(_get_value(state_dict, "n_bits", "quantizer.n_bits"))
    group_size = int(_get_value(state_dict, "group_size", "quantizer.group_size"))

    pairs = _stack_if_numbered(state_dict, "pairs_grouped").to(device=device, dtype=torch.short)
    theta = _stack_if_numbered(state_dict, "angles_grouped").to(device=device, dtype=torch.float32)

    channel_scales_opt = state_dict["channel_scales"].to(device=device, dtype=torch.float32)
    if channel_scales_opt.ndim == 1:
        channel_scales_opt = channel_scales_opt.unsqueeze(0)

    rotated = scaled_pairwise_rotation(weight * channel_scales_opt, pairs, theta, None, group_size)

    n_groups = in_features // group_size
    scales_flat = state_dict["quantizer.scale"].to(device=device, dtype=torch.float32).reshape(-1, 1)
    zp_flat = state_dict["quantizer.zero_point_float"].to(device=device, dtype=torch.float32).reshape(-1, 1)
    zero_points = torch.clamp(-torch.round(zp_flat), 0, (1 << bits) - 1)

    quantized = (
        torch.clamp(
            torch.round(rotated.reshape(-1, group_size) / scales_flat) + zero_points,
            0,
            (1 << bits) - 1,
        )
        .to(torch.int32)
        .reshape(out_features, in_features)
    )

    channel_scales = (1.0 / channel_scales_opt).to(torch.float16).cpu()
    if channel_scales.ndim == 1:
        channel_scales = channel_scales.unsqueeze(0)

    buffers: dict[str, torch.Tensor] = {
        "qweight": _pack_awq(quantized.T.contiguous()).cpu(),
        "qzeros": _pack_awq(zero_points.to(torch.int32).reshape(out_features, n_groups).T.contiguous()).cpu(),
        "scales": scales_flat.reshape(out_features, n_groups).T.contiguous().to(torch.float16).cpu(),
        "theta": theta.to(torch.float16).cpu(),
        "pairs": pairs.cpu(),
        "channel_scales": channel_scales,
    }
    if "bias" in state_dict and state_dict["bias"] is not None:
        buffers["bias"] = state_dict["bias"].to(torch.float16).cpu()

    return buffers, bits, group_size, int(theta.shape[0])


@torch.no_grad()
def _convert_real(model: torch.nn.Module, result_dir: Path) -> tuple[int, dict[str, Any]]:
    from paroquant.inference.backends.transformers.modules import RotateQuantizedLinear

    blocks = _get_blocks(model)
    count = 0
    bits = group_size = krot = 0

    for layer_idx, layer in enumerate(tqdm(blocks, desc="Quantizing")):
        for name, module in get_named_linears(layer).items():
            pt_file = result_dir / f"{layer_idx}.{name}.pt"
            if not pt_file.exists():
                continue

            sd = torch.load(pt_file, map_location="cpu", weights_only=False)
            buffers, bits, group_size, krot = _quantize_layer(sd, device="cuda")

            rl = RotateQuantizedLinear(
                module.in_features,
                module.out_features,
                bias=module.bias is not None,
                group_size=group_size,
                bits=bits,
                krot=krot,
            )
            rl.load_state_dict(buffers, strict=False)
            set_module_by_name(layer, name, rl)
            count += 1

        layer.cpu()
        torch.cuda.empty_cache()

    quant_config: dict[str, Any] = {
        "quant_method": "paroquant",
        "bits": bits,
        "group_size": group_size,
        "krot": krot,
    }
    return count, quant_config


@torch.no_grad()
def main() -> None:
    parser = ArgumentParser()
    parser.add_argument("--model", type=str, required=True)
    parser.add_argument("--result-dir", type=str, required=True)
    parser.add_argument("--output-path", type=str, required=True)
    parser.add_argument("--mode", choices=["real", "pseudo"], default="real")
    args = parser.parse_args()

    if not torch.cuda.is_available():
        raise RuntimeError("CUDA required.")

    result_dir = Path(args.result_dir)
    if not result_dir.is_dir():
        raise FileNotFoundError(f"Result directory not found: {result_dir}")

    model = _load_model(args.model, device_map="cpu" if args.mode == "real" else "cuda")

    if args.mode == "pseudo":
        count = _convert_pseudo(model, result_dir)
    else:
        count, quant_config = _convert_real(model, result_dir)
        args_json = result_dir / "args.json"
        if args_json.exists():
            skipped = json.loads(args_json.read_text()).get("skipped_modules", [])
            if skipped:
                quant_config["modules_to_not_convert"] = skipped
        model.config.quantization_config = quant_config

    if count == 0:
        raise RuntimeError(f"No checkpoint files matched in {result_dir}")

    output_path = Path(args.output_path)
    output_path.mkdir(parents=True, exist_ok=True)
    model.save_pretrained(output_path)
    AutoTokenizer.from_pretrained(args.model).save_pretrained(output_path)

    print(f"Converted {count} layers ({args.mode}) → {output_path}")


if __name__ == "__main__":
    main()



import argparse
import asyncio
import time

from paroquant.inference import GenerationParams, build_prompt, create_generator


async def bench(gen, prompt: str, max_tokens: int, warmup: int, runs: int):
    params = GenerationParams(max_tokens=max_tokens, temperature=0.0)
    formatted = build_prompt(gen.tokenizer, [{"role": "user", "content": prompt}])

    for _ in range(warmup):
        await gen.generate(formatted, params)

    times, tokens = [], []
    for _ in range(runs):
        start = time.perf_counter()
        result = await gen.generate(formatted, params)
        elapsed = time.perf_counter() - start
        times.append(elapsed)
        tokens.append(result.stats.num_tokens)

    avg_tokens = sum(tokens) / len(tokens)
    avg_time = sum(times) / len(times)
    avg_tps = avg_tokens / avg_time if avg_time > 0 else 0
    return avg_tokens, avg_time, avg_tps


async def main():
    parser = argparse.ArgumentParser(description="Benchmark generation throughput")
    parser.add_argument("--model", type=str, required=True)
    parser.add_argument("--backend", type=str, default="auto", choices=["auto", "mlx", "vllm", "transformers"])
    parser.add_argument("--max-tokens", type=int, default=128)
    parser.add_argument("--warmup", type=int, default=2)
    parser.add_argument("--runs", type=int, default=5)
    parser.add_argument("--prompt", type=str, default="Write a short essay about the history of computing.")
    args = parser.parse_args()

    print(f"Loading {args.model} with {args.backend} backend...")
    start = time.perf_counter()
    gen = create_generator(args.backend, args.model)
    load_time = time.perf_counter() - start
    print(f"Loaded in {load_time:.1f}s")

    print(f"Benchmarking: {args.warmup} warmup + {args.runs} runs, {args.max_tokens} tokens each")
    avg_tokens, avg_time, avg_tps = await bench(gen, args.prompt, args.max_tokens, args.warmup, args.runs)
    print(f"  Avg tokens: {avg_tokens:.0f}")
    print(f"  Avg time:   {avg_time:.2f}s")
    print(f"  Throughput:  {avg_tps:.1f} tok/s")


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import importlib
import time
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Callable
from dataclasses import dataclass


@dataclass
class GenerationParams:
    max_tokens: int = 512
    temperature: float = 1.0
    top_p: float = 1.0
    top_k: int = -1
    repetition_penalty: float = 1.0


@dataclass
class GenerationStats:
    num_tokens: int
    latency: float
    ttft: float | None
    tps: float


@dataclass
class GenerationResult:
    backend: str
    prompt: str
    output_text: str
    stats: GenerationStats


class BaseGenerator(ABC):
    backend: str = "unknown"

    @abstractmethod
    async def stream_generate(
        self,
        prompt: str,
        params: GenerationParams,
    ) -> AsyncIterator[str]:
        """Yield text chunks one at a time."""
        ...

    async def generate(
        self,
        prompt: str,
        params: GenerationParams,
        on_text: Callable[[str], None] | None = None,
    ) -> GenerationResult:
        start = time.perf_counter()
        first_token_time = None
        chunks: list[str] = []

        async for text in self.stream_generate(prompt, params):
            if first_token_time is None:
                first_token_time = time.perf_counter()
            chunks.append(text)
            if on_text:
                on_text(text)

        end = time.perf_counter()
        output_text = "".join(chunks)
        gen_time = end - (first_token_time or start)
        num_tokens = self._count_tokens(output_text)

        return GenerationResult(
            backend=self.backend,
            prompt=prompt,
            output_text=output_text,
            stats=GenerationStats(
                num_tokens=num_tokens,
                latency=end - start,
                ttft=(first_token_time - start) if first_token_time else None,
                tps=num_tokens / gen_time if gen_time > 0 else 0.0,
            ),
        )

    def _count_tokens(self, text: str) -> int:
        return len(self.tokenizer.encode(text, add_special_tokens=False))

    async def close(self) -> None:
        pass


def build_prompt(tokenizer, messages: list[dict[str, str]], enable_thinking: bool = False) -> str:
    """Apply chat template to messages, with fallback for tokenizers without templates."""
    if tokenizer.chat_template is None:
        return messages[-1]["content"]
    try:
        return tokenizer.apply_chat_template(
            messages,
            tokenize=False,
            add_generation_prompt=True,
            enable_thinking=enable_thinking,
        )
    except TypeError:
        return tokenizer.apply_chat_template(
            messages,
            tokenize=False,
            add_generation_prompt=True,
        )


_BACKENDS = {
    "transformers": ("paroquant.inference.backends.transformers", "TransformersGenerator", "paroquant[transformers]"),
    "vllm": ("paroquant.inference.backends.vllm", "VllmGenerator", "paroquant[vllm]"),
    "mlx": ("paroquant.inference.backends.mlx", "MlxGenerator", "paroquant[mlx]"),
}


def detect_backend() -> str:
    import platform

    if platform.system() == "Darwin":
        try:
            import mlx.core  # noqa: F401

            return "mlx"
        except ImportError:
            pass
    try:
        import torch

        if torch.cuda.is_available():
            try:
                import vllm  # noqa: F401

                return "vllm"
            except ImportError:
                return "transformers"
    except ImportError:
        pass
    raise RuntimeError("No backend available. Install paroquant[mlx], paroquant[vllm], or paroquant[transformers].")


def create_generator(backend: str, model: str, **kwargs) -> BaseGenerator:
    """Factory that instantiates the right backend by name."""
    key = backend.lower()
    if key == "auto":
        key = detect_backend()

    if key not in _BACKENDS:
        raise ValueError(f"Unknown backend: {backend!r}. Choose from: {', '.join(_BACKENDS)}")

    module_path, class_name, install_hint = _BACKENDS[key]
    try:
        cls = getattr(importlib.import_module(module_path), class_name)
    except (ImportError, ModuleNotFoundError) as e:
        raise ImportError(f'Backend {backend!r} requires: pip install "{install_hint}"') from e

    return cls(model=model, **kwargs)



from __future__ import annotations

import torch


def _align_shape(pair, angle, grp_size=128, inc_mask=False):
    """Align rotation pairs into fixed-size groups with dummy padding."""
    assert pair.size(0) == angle.size(0)
    assert pair.size(1) == 2

    grp = 0
    pair_ptr = 0

    pair_groups = []
    angle_groups = []
    mask_groups = []
    while True:
        if pair_ptr >= pair.size(0):
            break
        s = torch.zeros((grp_size), dtype=torch.int32)
        count = 0
        temp_pairs = torch.zeros((grp_size // 2, 2), dtype=torch.int32, device=pair.device)
        temp_angle = torch.zeros((grp_size // 2), dtype=torch.float, device=angle.device)
        temp_mask = torch.zeros((grp_size // 2), dtype=torch.int32, device=angle.device)
        while count < grp_size // 2:
            if (
                pair_ptr < pair.size(0)
                and pair[pair_ptr, 0] - grp * grp_size < grp_size
                and pair[pair_ptr, 1] - grp * grp_size < grp_size
            ):
                temp_pairs[count, :] = pair[pair_ptr, :]
                temp_angle[count] = angle[pair_ptr]
                if s[pair[pair_ptr, 0] % grp_size] == 1 or s[pair[pair_ptr, 1] % grp_size] == 1:
                    raise ValueError("illegal pair")
                s[pair[pair_ptr, :] % grp_size] = 1
                pair_ptr += 1
            else:
                t_pair = torch.tensor([-1, -1])
                for i in range(grp_size):
                    if s[i] == 0:
                        t_pair[0] = i
                        s[i] = 1
                        break
                for i in range(grp_size):
                    if s[i] == 0:
                        t_pair[1] = i
                        s[i] = 1
                        break
                if t_pair[0] == -1 or t_pair[1] == -1:
                    raise ValueError("can't find a dummy pair")
                temp_pairs[count, :] = t_pair
                temp_angle[count] = float(0)
                temp_mask[count] = 1
            count += 1
        grp += 1
        pair_groups.append(temp_pairs)
        angle_groups.append(temp_angle)
        mask_groups.append(temp_mask)

    rotation_pairs = torch.cat(pair_groups, dim=0).view(-1).contiguous()
    rotation_pairs = rotation_pairs % grp_size
    angles = torch.cat(angle_groups, dim=0)
    if inc_mask:
        masks = torch.cat(mask_groups, dim=0)
        return rotation_pairs, angles, masks
    return rotation_pairs, angles


def transform_to_kernel_data(
    pairs_group: list[torch.Tensor],
    angles_group: list[torch.Tensor],
    group_size: int = 128,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Convert optimization pair/angle lists to kernel-ready format with masks."""
    assert len(pairs_group) == len(angles_group)
    pairs, angles, masks = [], [], []
    for i in range(len(pairs_group)):
        pair, angle, mask = _align_shape(pairs_group[i], angles_group[i], grp_size=group_size, inc_mask=True)
        pairs.append(pair)
        angles.append(angle)
        masks.append(mask)

    return (
        torch.stack(pairs, dim=0).to(dtype=torch.short),
        torch.stack(angles, dim=0).to(dtype=torch.float),
        torch.stack(masks, dim=0).to(dtype=torch.bool),
    )



from __future__ import annotations

from collections.abc import AsyncIterator
from threading import Thread

import torch
from transformers import (
    AutoModelForCausalLM,
    AutoModelForImageTextToText,
    AutoTokenizer,
    TextIteratorStreamer,
)

import paroquant.inference.backends.transformers.quantizer  # noqa: F401 — registers HfQuantizer
from paroquant.inference.base import BaseGenerator, GenerationParams

_LOAD_KWARGS = dict(
    torch_dtype=torch.float16,
    low_cpu_mem_usage=True,
    attn_implementation="sdpa",
    device_map="cuda",
)


@torch.no_grad()
def _load_model(path: str):
    for auto_cls in (AutoModelForImageTextToText, AutoModelForCausalLM):
        try:
            model = auto_cls.from_pretrained(path, **_LOAD_KWARGS)
            model.eval()
            return model
        except (ValueError, KeyError):
            continue
    raise ValueError(f"Failed to load model from {path}")


class TransformersGenerator(BaseGenerator):
    backend = "transformers"

    def __init__(self, model: str):
        self.model = _load_model(model)
        self.tokenizer = AutoTokenizer.from_pretrained(model)

    async def stream_generate(self, prompt: str, params: GenerationParams) -> AsyncIterator[str]:
        inputs = self.tokenizer(prompt, return_tensors="pt").to(self.model.device)

        streamer = TextIteratorStreamer(self.tokenizer, skip_prompt=True, skip_special_tokens=True)
        thread = Thread(
            target=self.model.generate,
            kwargs={
                **inputs,
                "max_new_tokens": params.max_tokens,
                "do_sample": params.temperature > 0,
                "repetition_penalty": params.repetition_penalty if params.repetition_penalty > 1.0 else None,
                "temperature": params.temperature if params.temperature > 0 else None,
                "top_k": params.top_k if params.top_k > 0 else None,
                "top_p": params.top_p if params.top_p < 1.0 else None,
                "streamer": streamer,
            },
        )
        thread.start()

        for text in streamer:
            if text:
                yield text

        thread.join()



from __future__ import annotations

import torch
import torch.nn as nn
from torch.utils.checkpoint import checkpoint

from .quant import clamp_ste, round_ste


def _calc_scales_and_zero_points(
    weight: torch.Tensor, group_size: int, qmin: int, qmax: int
) -> tuple[torch.Tensor, torch.Tensor]:
    assert weight.dim() == 2, weight.shape
    if weight.dtype != torch.float32:
        weight = weight.float()
    x = weight.reshape(-1, group_size)
    min_val = x.amin(dim=1, keepdim=True)
    max_val = x.amax(dim=1, keepdim=True)
    scale = clamp_ste(max_val - min_val, min=1e-5) / qmax
    zero_point = min_val / scale

    assert torch.isnan(zero_point).sum() == 0, zero_point

    return scale, zero_point


# Adapted from EfficientQAT
class UniformAffineQuantizer(nn.Module):
    def __init__(
        self,
        weight: torch.Tensor,
        n_bits: int,
        group_size: int,
    ):
        super().__init__()
        # self.n_bits = n_bits
        if not isinstance(n_bits, torch.Tensor):
            n_bits = torch.tensor(n_bits)
        if not isinstance(group_size, torch.Tensor):
            group_size = torch.tensor(group_size)
        self.register_buffer("n_bits", n_bits)
        self.register_buffer("group_size", group_size)
        assert weight.shape[-1] % group_size == 0

        scale, zero_point_float = _calc_scales_and_zero_points(weight, group_size, self.qmin, self.qmax)
        self.scale = nn.Parameter(scale)
        self.zero_point_float = nn.Parameter(zero_point_float)

        self.enable_checkpoint = False

    @property
    def qmin(self) -> int:
        return 0

    @property
    def qmax(self) -> torch.Tensor:
        return 2**self.n_bits - 1

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if torch.is_grad_enabled() and self.enable_checkpoint:
            return checkpoint(
                self.pseudo_quantize,
                x,
                self.n_bits,
                self.group_size,
                self.scale,
                self.zero_point_float,
                use_reentrant=False,
            )
        else:
            return self.pseudo_quantize(
                x,
                self.n_bits,
                self.group_size,
                self.scale,
                self.zero_point_float,
            )

    def optim_params(self) -> list[nn.Parameter]:
        return [self.scale, self.zero_point_float]

    def set_optim_enabled(self, enabled: bool):
        for param in self.optim_params():
            param.requires_grad = enabled

    @staticmethod
    def pseudo_quantize(
        x: torch.Tensor,
        n_bits: int | torch.Tensor,
        group_size: int | torch.Tensor,
        scale: torch.Tensor | None = None,
        zero_point: torch.Tensor | None = None,
    ) -> torch.Tensor:
        dtype = x.dtype
        if x.dtype != torch.float32:
            x = x.float()
        assert torch.isnan(x).sum() == 0, x

        qmin, qmax = 0, 2**n_bits - 1
        if scale is None or zero_point is None:
            scale, zero_point = _calc_scales_and_zero_points(x, group_size, qmin, qmax)

        scale = clamp_ste(scale, min=1e-5, max=1e5)
        round_zero_point = clamp_ste(-round_ste(zero_point), qmin, qmax)
        dim1, dim2 = x.shape
        x = x.reshape(-1, group_size)
        x_int = round_ste(x / scale)
        assert torch.isnan(x_int).sum() == 0, (x_int, scale.min(), scale.max())
        x_int = x_int + round_zero_point
        x_int = clamp_ste(x_int, qmin, qmax)
        x_dequant = x_int
        x_dequant = x_dequant - round_zero_point
        x_dequant = x_dequant * scale
        x_dequant = x_dequant.reshape(dim1, dim2)
        assert torch.isnan(x_dequant).sum() == 0, x_dequant
        assert torch.isinf(x_dequant).sum() == 0, x_dequant
        return x_dequant.to(dtype)



from __future__ import annotations

import logging
from pathlib import Path

import mlx.core as mx
import mlx.nn as mlx_nn
import numpy as np

from .modules import RotateQuantizedLinear

logger = logging.getLogger(__name__)

_BITS = 4
_PF = 32 // _BITS
_MASK = (1 << _BITS) - 1
_SHIFTS = np.arange(0, 32, _BITS, dtype=np.int64)
_INV_REORDER = np.array([0, 4, 1, 5, 2, 6, 3, 7])


def _unpack_and_reorder(packed: np.ndarray) -> np.ndarray:
    """Unpack AutoAWQ int32 → raw uint8, undoing the [0,2,4,6,1,3,5,7] reorder."""
    raw = ((packed.astype(np.int64)[:, :, None] >> _SHIFTS) & _MASK).astype(np.uint8)
    return raw[:, :, _INV_REORDER].reshape(packed.shape[0], -1)


def _pack_mlx(w: np.ndarray) -> np.ndarray:
    """Pack raw uint8 values into uint32 (MLX sequential layout)."""
    w = w.reshape(w.shape[0], -1, _PF)
    p = w[:, :, 0].astype(np.uint32)
    for i in range(1, _PF):
        p |= w[:, :, i].astype(np.uint32) << (_BITS * i)
    return p


def _convert_autoawq(weights: dict, group_size: int) -> dict:
    """Convert AutoAWQ int32 checkpoint to MLX quantized format."""
    prefixes = {
        k.removesuffix("qweight")
        for k in weights
        if k.endswith(".qweight") and f"{k[:-len('qweight')]}theta" in weights
    }
    if not prefixes:
        return weights

    out = {}
    for key, val in weights.items():
        pfx = next((p for p in prefixes if key.startswith(p)), None)
        if pfx is None:
            out[key] = val
            continue

        suffix = key[len(pfx) :]
        if suffix == "qweight":
            out[f"{pfx}weight"] = mx.array(_pack_mlx(_unpack_and_reorder(np.array(val)).T))
        elif suffix == "scales":
            out[f"{pfx}scales"] = mx.array(np.array(val).T.copy())
        elif suffix == "qzeros":
            zeros = _unpack_and_reorder(np.array(val)).astype(np.float32)
            scales = np.array(weights[f"{pfx}scales"]).astype(np.float32)
            out[f"{pfx}biases"] = mx.array((-scales * zeros).T.copy().astype(np.float16))
        elif suffix in ("theta", "pairs", "channel_scales", "bias"):
            out[key] = val.reshape(1, -1) if suffix == "channel_scales" and val.ndim == 1 else val

    return out


def _create_model(config: dict, is_vlm: bool):
    """Instantiate model from config, dispatching to mlx_vlm or mlx_lm."""
    if is_vlm:
        from mlx_vlm.utils import get_model_and_args, update_module_configs

        mod, _ = get_model_and_args(config=config)
        cfg = mod.ModelConfig.from_dict(config)
        cfg = update_module_configs(cfg, mod, config, ["text", "vision", "perceiver", "projector", "audio"])
        return mod.Model(cfg)

    from mlx_lm.utils import _get_classes

    model_class, model_args_class = _get_classes(config)
    return model_class(model_args_class.from_dict(config))


def _load_processor(local_dir: Path, model, is_vlm: bool):
    """Load tokenizer (LLM) or processor (VLM)."""
    if not is_vlm:
        from mlx_lm.utils import load_tokenizer

        return load_tokenizer(local_dir)

    from mlx_vlm.utils import load_image_processor, load_processor

    try:
        processor = load_processor(local_dir, trust_remote_code=True)
        image_processor = load_image_processor(local_dir)
        if image_processor is not None:
            processor.image_processor = image_processor
        return processor
    except (ImportError, TypeError, AttributeError):
        pass

    from mlx_vlm.utils import load_tokenizer

    processor = load_tokenizer(local_dir)
    if not hasattr(processor, "stopping_criteria"):
        eos = getattr(model.config, "eos_token_id", None)
        eos_set = set(eos) if isinstance(eos, list) else {eos} if eos is not None else set()
        processor.stopping_criteria = lambda token: int(token) in eos_set
    return processor


def _get_module(root, path: str):
    node = root
    for part in path.split("."):
        node = node[int(part)] if isinstance(node, (list, tuple)) else getattr(node, part)
    return node


def _set_module(root, path: str, value):
    parts = path.split(".")
    parent = _get_module(root, ".".join(parts[:-1])) if len(parts) > 1 else root
    if isinstance(parent, list):
        parent[int(parts[-1])] = value
    else:
        setattr(parent, parts[-1], value)


def _patch_rotation_layers(model, weights: dict, bits: int, group_size: int):
    """Replace quantized linear layers that have rotation params with RotateQuantizedLinear."""
    for prefix in sorted(k.removesuffix(".theta") for k in weights if k.endswith(".theta")):
        original = _get_module(model, prefix)
        _set_module(
            model,
            prefix,
            RotateQuantizedLinear(
                input_dims=original.weight.shape[-1],
                output_dims=original.weight.shape[0],
                bias="bias" in original,
                group_size=group_size,
                bits=bits,
                krot=weights[f"{prefix}.theta"].shape[0],
            ),
        )


def _is_io_layer(path, module):
    """Predicate for mlx_nn.quantize: only quantize embed_tokens/lm_head natively."""
    return hasattr(module, "to_quantized") and (path.endswith("embed_tokens") or path.endswith("lm_head"))


def load(model_path: str, lazy: bool = False, force_text: bool = False) -> tuple:
    """Load a ParoQuant model for MLX. Returns (model, processor, is_vlm)."""
    from huggingface_hub import snapshot_download

    local_dir = Path(model_path)
    if not local_dir.is_dir():
        local_dir = Path(snapshot_download(model_path))

    try:
        from mlx_vlm.utils import load_config
    except ImportError:
        from mlx_lm.utils import load_config
    config = load_config(local_dir)

    paro = config.get("quantization_config", {})
    group_size = int(paro.get("group_size", 128))
    bits = int(paro.get("bits", 4))
    is_vlm = "vision_config" in config and not force_text

    model = _create_model(config, is_vlm)

    weights = {}
    for sf in sorted(local_dir.glob("*.safetensors")):
        weights.update(mx.load(str(sf)))
    if any(k.endswith(".qweight") for k in weights):
        weights = _convert_autoawq(weights, group_size)
    if hasattr(model, "sanitize"):
        weights = model.sanitize(weights)

    _patch_rotation_layers(model, weights, bits, group_size)
    model.load_weights(list(weights.items()), strict=False)
    mlx_nn.quantize(model, group_size, bits, mode="affine", class_predicate=_is_io_layer)

    if not lazy:
        mx.eval(model.parameters())
    model.eval()

    processor = _load_processor(local_dir, model, is_vlm)
    logger.info("Loaded %s model from %s (VLM=%s).", config.get("model_type", "unknown"), model_path, is_vlm)

    return model, processor, is_vlm



from __future__ import annotations

import gc
import logging
import math
import random
import warnings
from typing import TypeVar

import torch
import torch.nn as nn
from datasets import load_dataset
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer


def get_blocks(model: nn.Module) -> nn.ModuleList:
    model_class_name = model.__class__.__name__
    if model_class_name in (
        "Qwen2ForCausalLM",
        "Qwen3ForCausalLM",
        "Qwen3MoeForCausalLM",
        "LlamaForCausalLM",
    ):
        m = model.model
    else:
        raise NotImplementedError(type(model))

    return m.layers


_Linear_T = TypeVar("Linear", bound=nn.Module)


def get_named_linears(module: nn.Module, subclass: type[_Linear_T] = nn.Linear) -> dict[str, _Linear_T]:
    return {name: m for name, m in module.named_modules() if isinstance(m, subclass)}


def get_module_by_name(module, module_name):
    for name, m in module.named_modules():
        if name == module_name:
            return m
    return None


def set_module_by_name(layer, name, new_module):
    levels = name.split(".")
    if len(levels) > 1:
        mod_ = layer
        for l_idx in range(len(levels) - 1):
            if levels[l_idx].isdigit():
                mod_ = mod_[int(levels[l_idx])]
            else:
                mod_ = getattr(mod_, levels[l_idx])
        setattr(mod_, levels[-1], new_module)
    else:
        setattr(layer, name, new_module)


def load_model(model_path: str, device_map: str | None = None, dtype=torch.float32, **kwargs) -> nn.Module:
    model = AutoModelForCausalLM.from_pretrained(model_path, device_map=device_map, torch_dtype=dtype, **kwargs)
    return model


def load_tokenizer(model_path: str, **kwargs) -> AutoTokenizer:
    tokenizer = AutoTokenizer.from_pretrained(model_path, **kwargs)
    tokenizer.pad_token = tokenizer.eos_token
    return tokenizer


def move_embed(model, device):
    model_class_name = model.__class__.__name__
    if model_class_name in (
        "Qwen2ForCausalLM",
        "Qwen3ForCausalLM",
        "Qwen3MoeForCausalLM",
        "LlamaForCausalLM",
    ):
        model.model.embed_tokens = model.model.embed_tokens.to(device)
        model.model.rotary_emb = model.model.rotary_emb.to(device)
    else:
        raise NotImplementedError(type(model))


def empty_cache():
    gc.collect()
    torch.cuda.empty_cache()


def get_mixed_calib_dataset(
    datasets: list[str],
    *,
    tokenizer,
    n_samples: int,
    block_size: int,
    seed: int,
    split: str,
) -> list[torch.Tensor]:
    per_dataset_len = n_samples // len(datasets)
    results = []
    for i, dataset in enumerate(datasets):
        dataset_samples = per_dataset_len if i < len(datasets) - 1 else n_samples - len(results)
        results.extend(
            get_calib_dataset(
                data=dataset,
                tokenizer=tokenizer,
                n_samples=dataset_samples,
                block_size=block_size,
                seed=seed,
                split=split,
            )
        )
    assert len(results) == n_samples, f"Expected {n_samples} samples, got {len(results)}"

    rand = random.Random(seed)
    rand.shuffle(results)

    return results


# Adapted from awq-llm
def get_calib_dataset(
    data="pileval",
    *,
    tokenizer,
    n_samples: int,
    block_size: int,
    seed: int,
    split: str,
) -> list[torch.Tensor]:
    if data == "pileval":
        if split != "validation":
            warnings.warn("The split argument is ignored when data is 'pileval'.")
        dataset = load_dataset("mit-han-lab/pile-val-backup", split="validation")
        dataset = dataset.shuffle(seed=seed)
    elif data == "wikitext2":
        dataset = load_dataset("wikitext", "wikitext-2-raw-v1", split=split)
        dataset = dataset.shuffle(seed=seed)
    elif data == "c4":
        if split == "train":
            dataset = load_dataset(
                "allenai/c4",
                data_files={"train": "en/c4-train.00000-of-01024.json.gz"},
                split=split,
            )
        elif split == "validation":
            dataset = load_dataset(
                "allenai/c4",
                data_files={"validation": "en/c4-validation.00001-of-00008.json.gz"},
                split=split,
            )
        dataset = dataset.shuffle(seed=seed)
    elif data == "redpajama":
        test_split, val_split = 0.2, 0.1
        dataset = load_dataset(
            "liang2kl/RedPajama-Data-1T-Sample-Backup",
            split="train",
            trust_remote_code=True,
        )
        dataset = dataset.shuffle(seed=seed)
        test_size = int(len(dataset) * test_split)
        val_size = int(len(dataset) * val_split)
        train_size = len(dataset) - test_size - val_size
        if split == "test":
            dataset = dataset.select(range(len(dataset) - test_size, len(dataset)))
        elif split == "validation":
            dataset = dataset.select(range(len(dataset) - test_size - val_size, len(dataset) - test_size))
        elif split == "train":
            dataset = dataset.select(range(0, train_size))
        else:
            raise ValueError(f"Invalid split: {split}")
    else:
        raise NotImplementedError

    samples = []
    total_len = 0
    for row in dataset:
        line = row["text"]
        line = line.strip()
        line_encoded = tokenizer.encode(line)
        if len(line_encoded) > block_size:
            continue
        sample = torch.tensor([line_encoded])
        if sample.numel() == 0:
            continue
        samples.append(sample)
        total_len += len(line_encoded)
        if total_len >= n_samples * block_size:
            break
    samples = torch.cat(samples, dim=1).squeeze(0)
    n_split = min(samples.shape[0] // block_size, n_samples)

    return [samples[i * block_size : (i + 1) * block_size] for i in range(n_split)]


@torch.no_grad()
def catch_first_layer_input(
    model: nn.Module,
    layers: nn.ModuleList,
    samples: torch.Tensor,
    batch_size: int | None,
) -> tuple[torch.Tensor, dict]:
    layer_kwargs = {}
    batched = batch_size is not None
    inps: list[torch.Tensor] = []

    class Catcher(nn.Module):
        def __init__(self, module):
            super().__init__()
            # Bypass __setattr__ of nn.Module
            object.__setattr__(self, "module", module)

        def forward(self, inp, **kwargs):
            inps.append(inp)
            if len(layer_kwargs) == 0:
                layer_kwargs.update(kwargs)
            raise ValueError

        def __getattr__(self, name):
            return getattr(self.module, name)

    layers[0] = Catcher(layers[0])
    batch_size = samples.shape[0] if not batched or batch_size <= 0 else batch_size
    num_batches = samples.shape[0] // batch_size
    samples_batch = samples.chunk(num_batches)
    for samples in samples_batch:
        try:
            model(samples.to(next(model.parameters()).device))
        except ValueError:
            pass
    layers[0] = layers[0].module
    if not batched:
        inps = inps[0]

    layer_kwargs["use_cache"] = False
    if "past_key_value" in layer_kwargs:
        del layer_kwargs["past_key_value"]
    if "past_key_values" in layer_kwargs:
        del layer_kwargs["past_key_values"]

    return inps, layer_kwargs


class CachedTensorShards:
    def __init__(
        self,
        batches: list[torch.Tensor],
        num_shards: int,
        *,
        target_device: torch.device,
        offload_device: torch.device = torch.device("cpu"),
    ):
        assert len(batches) % num_shards == 0
        if batches[0].device != offload_device:
            self.batches = [b.to(offload_device) for b in batches]
        else:
            self.batches = batches
        self.num_shards = num_shards
        self.current_shard: int = None
        self.cached_shard: list[torch.Tensor] = None
        self.target_device = target_device

    def _switch_shard(self, shard_index: int) -> None:
        if self.current_shard == shard_index:
            return
        self.current_shard = shard_index
        start, end = self._get_shard_range(shard_index)
        self.cached_shard = self.batches[start:end]
        self.cached_shard = [b.to(self.target_device) for b in self.cached_shard]

    def _get_shard_range(self, index: int) -> tuple[int, int]:
        if self.num_shards == 1:
            return 0, len(self.batches)
        shard_size = len(self.batches) // self.num_shards
        start = shard_size * index
        if index == self.num_shards - 1:
            end = len(self.batches)
        else:
            end = shard_size * (index + 1)
        return start, end

    def __getitem__(self, index: int) -> torch.Tensor:
        shard_len = len(self.batches) // self.num_shards
        shard_index = index // shard_len
        if self.current_shard != shard_index:
            self._switch_shard(shard_index)
        offset = index % shard_len
        return self.cached_shard[offset]

    def __iter__(self) -> "Iterator":
        return self.Iterator(self)

    def __len__(self) -> int:
        return len(self.batches)

    class Iterator:
        def __init__(self, batches: "CachedTensorShards"):
            self.batches = batches
            self.current_index = 0

        def __iter__(self):
            return self

        def __next__(self) -> torch.Tensor:
            if self.current_index >= len(self.batches):
                raise StopIteration
            result = self.batches[self.current_index]
            self.current_index += 1
            return result

        def __len__(self) -> int:
            return len(self.batches)


class CosineAnnealingParam:
    def __init__(self, start_value: float, end_value: float, T_max: int):
        """
        Args:
            start_value (float): The initial value (equivalent to eta_max).
            end_value (float): The final value (equivalent to eta_min).
            T_max (int): Maximum number of steps.
        """
        self.start_value = start_value
        self.end_value = end_value
        self.T_max = T_max
        self._step = -1

    def step(self) -> float:
        self._step += 1

        if self._step >= self.T_max:
            return self.end_value

        cos_val = math.cos(math.pi * self._step / self.T_max)
        return self.end_value + (self.start_value - self.end_value) * (1 + cos_val) / 2


class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    handler = TqdmLoggingHandler()
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    return logger


logger = get_logger("ParoQuant")



from __future__ import annotations

import argparse
import json
import logging
import os
import re
import tempfile
import time
import warnings

from qwen_agent.agents import Assistant
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.theme import Theme

_SPECIAL_RE = re.compile(r"<\|[^|]+\|>")
_THINK_RE = re.compile(r"^.*?</think>\n?", re.DOTALL)
_THINKING_LINES = 4

SYSTEM_PROMPT = """\
You are ParoQuant Agent, a helpful AI assistant powered by a local quantized model.
You have access to tools that let you interact with the real world:
- Get the current time in any timezone
- Fetch and read web pages
- Read, write, and list files in the workspace directory
Use tools when the user's request requires real-world information or actions.
Be concise in your final answers."""


def _suppress_library_noise():
    os.environ["PYTHONWARNINGS"] = "ignore"
    warnings.filterwarnings("ignore")
    logging.disable(logging.CRITICAL)


def _make_agent(server: str, model: str, workspace: str, max_tokens: int = 8192) -> Assistant:
    llm_cfg = {
        "model": model,
        "model_type": "qwenvl_oai",
        "model_server": server,
        "api_key": os.environ.get("OPENAI_API_KEY", "EMPTY"),
        "generate_cfg": {
            "max_tokens": max_tokens,
            "timeout": 120,
            "max_retries": 1,
            "temperature": 0.7,
            "top_p": 0.8,
            "presence_penalty": 1.5,
            "extra_body": {
                "top_k": 20,
                "min_p": 0.0,
                "repetition_penalty": 1.0,
                "chat_template_kwargs": {"enable_thinking": False},
            },
        },
    }

    mcp_servers = {
        "time": {"command": "uvx", "args": ["mcp-server-time", "--local-timezone=US/Pacific"]},
        "fetch": {"command": "uvx", "args": ["mcp-server-fetch"]},
        "filesystem": {"command": "npx", "args": ["-y", "@modelcontextprotocol/server-filesystem", workspace]},
    }

    tools: list = []
    try:
        import mcp  # noqa: F401

        tools = [{"mcpServers": mcp_servers}]
    except ImportError:
        pass

    return Assistant(
        llm=llm_cfg,
        function_list=tools or None,
        name="ParoQuant Agent",
        description="Agent powered by a local ParoQuant quantized model with tool calling.",
        system_message=SYSTEM_PROMPT,
    )


# ---------------------------------------------------------------------------
# Response parsing helpers
# ---------------------------------------------------------------------------


def _clean(text: str) -> str:
    return _THINK_RE.sub("", _SPECIAL_RE.sub("", text)).strip()


def _get_thinking(text: str) -> str | None:
    if "<think>" in text and "</think>" not in text:
        return text.split("<think>", 1)[1]
    return None


def _friendly_tool_name(name: str) -> str:
    parts = name.split("-", 1)
    return parts[1] if len(parts) > 1 else name


def _format_tool_args(args_json: str) -> str:
    try:
        return ", ".join(f"{k}={v}" for k, v in json.loads(args_json).items())
    except (json.JSONDecodeError, TypeError, AttributeError):
        return args_json or ""


def _last_msg_with(msgs: list[dict], *, role: str = "", key: str = "") -> tuple[dict | None, int]:
    """Return the last message matching role/key and its index, or (None, -1)."""
    for i in reversed(range(len(msgs))):
        m = msgs[i]
        if not isinstance(m, dict):
            continue
        if role and m.get("role") != role:
            continue
        if key and not m.get(key):
            continue
        return m, i
    return None, -1


def _thinking_panel(width: int, text: str, elapsed: float) -> Panel:
    tail = "\n".join(text.splitlines()[-_THINKING_LINES:])
    return Panel(
        tail or "...",
        title=f"thinking ({elapsed:.1f}s)",
        border_style="dim",
        width=min(width, 80),
        height=_THINKING_LINES + 2,
    )


# ---------------------------------------------------------------------------
# Main loop
# ---------------------------------------------------------------------------


def main():
    _suppress_library_noise()

    parser = argparse.ArgumentParser(description="ParoQuant Agent")
    parser.add_argument("--server", type=str, default="http://127.0.0.1:8000/v1")
    parser.add_argument("--model", type=str, required=True)
    parser.add_argument("--workspace", type=str, default=None)
    parser.add_argument("--max-tokens", type=int, default=8192)
    args = parser.parse_args()

    workspace = args.workspace or tempfile.mkdtemp(prefix="paroquant-agent-")
    console = Console(theme=Theme({"hint": "dim", "tool": "cyan"}))

    console.print("[hint]Connecting to model server and initializing tools...[/hint]")
    agent = _make_agent(args.server, args.model, workspace, args.max_tokens)

    console.print("[hint]Warming up...[/hint]")
    warmup = _make_agent(args.server, args.model, workspace, max_tokens=16)
    for _ in warmup.run(messages=[{"role": "user", "content": "Hi"}]):
        pass
    del warmup

    console.clear()

    console.print(
        Panel.fit(
            f"[bold]ParoQuant Agent[/bold]\n"
            f"Model: [bold]{args.model}[/bold]\n"
            f"Server: [bold]{args.server}[/bold]\n\n"
            f"[bold]/clear[/bold] · [bold]/quit[/bold]",
            border_style="bright_blue",
        )
    )

    messages: list[dict] = []

    while True:
        try:
            user_input = console.input(">>> ").strip()
        except (KeyboardInterrupt, EOFError):
            console.print("\n[hint]Session closed.[/hint]")
            break
        if not user_input:
            continue
        if user_input.lower() in {"/quit", "quit", "/exit", "exit"}:
            break
        if user_input.lower() == "/clear":
            messages.clear()
            console.clear()
            console.print("[hint]Conversation history cleared.[/hint]\n")
            continue

        messages.append({"role": "user", "content": user_input})

        last_response: list[dict] = []
        final_content = ""
        turn_start = time.perf_counter()
        live: Live | None = None
        phase = "idle"
        seen_tool_idx = seen_resp_idx = -1
        tool_time = 0.0
        tool_timer_start: float | None = None

        try:
            for response in agent.run(messages=messages):
                if not isinstance(response, list):
                    continue
                last_response = response
                elapsed = time.perf_counter() - turn_start

                # Detect new tool call
                tc_msg, tc_idx = _last_msg_with(response, key="function_call")
                if tc_idx > seen_tool_idx and tc_idx == len(response) - 1:
                    seen_tool_idx = tc_idx
                    tool_timer_start = time.perf_counter()
                    phase = "tool"
                    continue

                # Detect new tool response
                _, resp_idx = _last_msg_with(response, role="function")
                if resp_idx > seen_resp_idx and resp_idx == len(response) - 1:
                    seen_resp_idx = resp_idx
                    if tool_timer_start is not None:
                        tool_time += time.perf_counter() - tool_timer_start
                        tool_timer_start = None
                    continue

                # Tool call completed — print summary and move on
                if phase == "tool" and tc_msg is not None and tc_idx != len(response) - 1:
                    if live:
                        live.stop()
                        live = None
                    fc = tc_msg["function_call"]
                    name = _friendly_tool_name(fc.get("name", ""))
                    console.print(f"  [tool]> {name}({_format_tool_args(fc.get('arguments', ''))})[/tool]")
                    resp_msg, _ = _last_msg_with(response, role="function")
                    if resp_msg:
                        text = resp_msg.get("content", "").replace("\n", " ")
                        if len(text) > 100:
                            text = text[:100] + "..."
                        console.print(f"    [hint]{text}[/hint]")
                    phase = "generating"

                # Extract latest assistant content
                asst_msg, _ = _last_msg_with(response, role="assistant")
                raw = (asst_msg or {}).get("content", "")
                if not raw:
                    continue

                # Show thinking box
                thinking = _get_thinking(raw)
                if thinking is not None:
                    if phase != "thinking":
                        if live:
                            live.stop()
                        live = Live(console=console, transient=True)
                        live.start()
                        phase = "thinking"
                    live.update(_thinking_panel(console.width, thinking, elapsed))
                    continue

                # Stream answer
                content = _clean(raw)
                if not content or content == final_content:
                    continue
                if phase != "answering":
                    if live:
                        live.stop()
                    live = Live(Markdown(content), console=console, vertical_overflow="visible")
                    live.start()
                    phase = "answering"
                else:
                    live.update(Markdown(content))
                final_content = content

        except Exception as e:
            if live:
                live.stop()
            console.print(f"[red]Error: {e}[/red]\n")
            messages.pop()
            continue

        if live:
            if final_content:
                live.update(Markdown(final_content))
            live.stop()

        if not final_content:
            asst_msg, _ = _last_msg_with(last_response, role="assistant")
            final_content = _clean((asst_msg or {}).get("content", ""))
            if final_content:
                console.print(Markdown(final_content))

        if tool_timer_start is not None:
            tool_time += time.perf_counter() - tool_timer_start

        total = time.perf_counter() - turn_start
        stats = []
        if tool_time > 0:
            stats.append(f"tools {tool_time:.1f}s")
        stats.append(f"{total:.1f}s total")
        console.print(f"  {' · '.join(stats)}", style="hint", highlight=False)

        if last_response:
            messages.extend(last_response)
        console.print()


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import torch
from torch.nn import Parameter
from vllm.logger import init_logger
from vllm.model_executor.layers.linear import LinearBase, LinearMethodBase, UnquantizedLinearMethod
from vllm.model_executor.layers.quantization import register_quantization_config
from vllm.model_executor.layers.quantization.awq_marlin import AWQMarlinLinearMethod
from vllm.model_executor.layers.quantization.base_config import QuantizationConfig
from vllm.model_executor.layers.quantization.utils.marlin_utils import (
    apply_awq_marlin_linear,
    check_marlin_supports_layer,
)
from vllm.model_executor.layers.quantization.utils.quant_utils import is_layer_skipped
from vllm.scalar_type import scalar_types
from vllm.transformers_utils.config import get_safetensors_params_metadata

import paroquant.kernels.cuda  # noqa: F401 — registers torch.ops.rotation.rotate

if TYPE_CHECKING:
    from vllm.model_executor.layers.quantization import QuantizationMethods

logger = init_logger(__name__)

_SHARD_INDEX = {"q": 0, "k": 1, "v": 2}
_QUANT_TYPE = {4: scalar_types.uint4}
_MARLIN_TILE_N = 64


def _rotation_weight_loader(
    param: Parameter,
    loaded_weight: torch.Tensor,
    loaded_shard_id: int | str | tuple | None = None,
) -> None:
    """Load per-projection rotation params into the partitioned param tensor.

    vLLM calls this with different shard_id types depending on the merge:
      None         → single projection, copy directly
      "q"/"k"/"v"  → QKV merge, map to partition index 0/1/2
      int          → gate/up merge, use as partition index
      tuple        → fused projections (e.g. Qwen3.5), copy to each index
    """
    if loaded_shard_id is None:
        target = param.data[0] if param.data.dim() > loaded_weight.dim() else param.data
        target.copy_(loaded_weight)
        return

    indices = (
        loaded_shard_id if isinstance(loaded_shard_id, tuple) else (_SHARD_INDEX.get(loaded_shard_id, loaded_shard_id),)
    )
    for idx in indices:
        param.data[idx].copy_(loaded_weight)


@register_quantization_config("paroquant")
class ParoQuantConfig(QuantizationConfig):
    def __init__(self, bits: int, group_size: int, krot: int, modules_to_not_convert: list[str] | None = None) -> None:
        super().__init__()
        if bits not in _QUANT_TYPE:
            raise ValueError(f"Unsupported bits={bits}. Supported: {list(_QUANT_TYPE)}")
        self.bits = bits
        self.group_size = group_size
        self.krot = krot
        self.pack_factor = 32 // bits
        self.quant_type = _QUANT_TYPE[bits]
        self.modules_to_not_convert = modules_to_not_convert or []

    def __repr__(self) -> str:
        return f"ParoQuantConfig(bits={self.bits}, group_size={self.group_size}, krot={self.krot})"

    @classmethod
    def get_name(cls) -> "QuantizationMethods":
        return "paroquant"

    @classmethod
    def get_supported_act_dtypes(cls) -> list[torch.dtype]:
        return [torch.half, torch.bfloat16]

    @classmethod
    def get_min_capability(cls) -> int:
        return 75

    @classmethod
    def get_config_filenames(cls) -> list[str]:
        return ["config.json"]

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> ParoQuantConfig:
        return cls(
            bits=cls.get_from_keys_or(config, ["bits"], 4),
            group_size=cls.get_from_keys_or(config, ["group_size"], 128),
            krot=cls.get_from_keys_or(config, ["krot"], 8),
            modules_to_not_convert=config.get("modules_to_not_convert"),
        )

    def maybe_update_config(self, model_name: str, revision: str | None = None):
        """Auto-detect unquantized layers from safetensors metadata."""
        if self.modules_to_not_convert:
            return

        from safetensors.torch import _TYPES as _SF_DTYPES

        unquant_dtypes = [torch.float16, torch.bfloat16, torch.float32]
        metadata = get_safetensors_params_metadata(model_name, revision=revision)

        # Only consider leaf modules (those with ".weight"), not containers
        # that have scalar FP16 params (e.g. A_log) which would false-match.
        leaf_modules = {k.rsplit(".", 1)[0] for k in metadata if k.endswith(".weight")}
        quant_modules: set[str] = {
            k.rsplit(".", 1)[0]
            for k, info in metadata.items()
            if (dt := info.get("dtype")) and _SF_DTYPES[dt] not in unquant_dtypes
        }
        # Strip to "layers.N..." suffix so vLLM's substr matching works
        # regardless of model nesting depth (safetensors may store
        # "model.language_model.layers.0.X" while vLLM uses
        # "language_model.model.layers.0.X").
        def _strip(name: str) -> str:
            name = name.removeprefix("model.")
            i = name.find("layers.")
            return name[i:] if i >= 0 else name

        self.modules_to_not_convert = [_strip(k) for k in leaf_modules - quant_modules]

    def get_quant_method(self, layer: torch.nn.Module, prefix: str) -> LinearMethodBase | None:
        if not isinstance(layer, LinearBase):
            return None
        if is_layer_skipped(prefix, self.modules_to_not_convert, self.packed_modules_mapping, skip_with_substr=True):
            return UnquantizedLinearMethod()
        if not check_marlin_supports_layer(layer, self.group_size):
            logger.warning_once(
                "Layer '%s' is not supported by Marlin. Falling back to unquantized.",
                prefix,
            )
            return UnquantizedLinearMethod()
        return ParoQuantLinearMethod(self)


class ParoQuantLinearMethod(AWQMarlinLinearMethod):
    """Per-projection rotation followed by AWQ-Marlin INT4 matmul."""

    def __init__(self, quant_config: ParoQuantConfig) -> None:
        super().__init__(quant_config)

    def create_weights(
        self,
        layer: torch.nn.Module,
        input_size_per_partition: int,
        output_partition_sizes: list[int],
        input_size: int,
        output_size: int,
        params_dtype: torch.dtype,
        **extra_weight_attrs,
    ) -> None:
        super().create_weights(
            layer,
            input_size_per_partition,
            output_partition_sizes,
            input_size,
            output_size,
            params_dtype,
            **extra_weight_attrs,
        )

        n_parts = len(output_partition_sizes)
        krot = self.quant_config.krot
        for name, shape, dtype in [
            ("theta", (n_parts, krot, input_size_per_partition // 2), torch.float16),
            ("pairs", (n_parts, krot, input_size_per_partition), torch.int16),
            ("channel_scales", (n_parts, 1, input_size_per_partition), torch.float16),
        ]:
            init_fn = torch.ones if name == "channel_scales" else torch.zeros
            p = Parameter(init_fn(shape, dtype=dtype), requires_grad=False)
            p.weight_loader = _rotation_weight_loader
            layer.register_parameter(name, p)

        layer.num_partitions = n_parts
        layer.output_partition_sizes = output_partition_sizes

    def _convert_partition(self, qw, sc, qz, k, out_n, num_groups):
        """AWQ→Marlin conversion for a single partition via the parent method."""
        # Pad output dim to Marlin tile boundary when needed.
        if out_n % _MARLIN_TILE_N != 0:
            pad = _MARLIN_TILE_N - out_n % _MARLIN_TILE_N
            pack = self.quant_config.pack_factor
            qw = torch.nn.functional.pad(qw, (0, pad // pack))
            sc = torch.nn.functional.pad(sc, (0, pad))
            qz = torch.nn.functional.pad(qz, (0, pad // pack))
            out_n += pad

        proxy = torch.nn.Module()
        proxy.register_parameter("qweight", Parameter(qw.contiguous(), requires_grad=False))
        proxy.register_parameter("scales", Parameter(sc.contiguous(), requires_grad=False))
        proxy.register_parameter("qzeros", Parameter(qz.contiguous(), requires_grad=False))
        proxy.input_size_per_partition = k
        proxy.output_size_per_partition = out_n
        proxy.num_groups = num_groups
        super().process_weights_after_loading(proxy)
        return proxy

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        n = layer.num_partitions

        if n == 1:
            super().process_weights_after_loading(layer)
        else:
            pack = self.quant_config.pack_factor
            sizes = layer.output_partition_sizes
            k = layer.input_size_per_partition

            qw = layer.qweight.data.split([s // pack for s in sizes], dim=1)
            sc = layer.scales.data.split(sizes, dim=1)
            qz = layer.qzeros.data.split([s // pack for s in sizes], dim=1)

            proxies = [self._convert_partition(qw[i], sc[i], qz[i], k, sizes[i], layer.num_groups) for i in range(n)]

            del layer.qweight, layer.scales, layer.qzeros
            layer.marlin_qweight = [p.qweight for p in proxies]
            layer.marlin_scales = [p.scales for p in proxies]
            layer.marlin_qzeros = [p.qzeros for p in proxies]
            layer.workspace = proxies[0].workspace
            layer.g_idx = proxies[0].g_idx
            layer.g_idx_sort_indices = proxies[0].g_idx_sort_indices
            layer.padded_partition_sizes = [p.output_size_per_partition for p in proxies]

        layer.rot_theta = layer.theta.data
        layer.rot_pairs = layer.pairs.data
        layer.rot_scales = layer.channel_scales.data
        del layer.theta, layer.pairs, layer.channel_scales

    def apply(self, layer: torch.nn.Module, x: torch.Tensor, bias: torch.Tensor | None = None) -> torch.Tensor:
        n = layer.num_partitions

        if n == 1:
            x = torch.ops.rotation.rotate(x, layer.rot_pairs[0], layer.rot_theta[0], layer.rot_scales[0])
            return super().apply(layer, x, bias)

        outputs = []
        for i in range(n):
            x_rot = torch.ops.rotation.rotate(x, layer.rot_pairs[i], layer.rot_theta[i], layer.rot_scales[i])
            out = apply_awq_marlin_linear(
                input=x_rot,
                weight=layer.marlin_qweight[i],
                weight_scale=layer.marlin_scales[i],
                weight_zp=layer.marlin_qzeros[i],
                g_idx=layer.g_idx,
                g_idx_sort_indices=layer.g_idx_sort_indices,
                workspace=layer.workspace,
                quant_type=self.quant_config.quant_type,
                output_size_per_partition=layer.padded_partition_sizes[i],
                input_size_per_partition=layer.input_size_per_partition,
                bias=None,
            )
            if layer.padded_partition_sizes[i] != layer.output_partition_sizes[i]:
                out = out[..., :layer.output_partition_sizes[i]]
            outputs.append(out)

        result = torch.cat(outputs, dim=-1)
        if bias is not None:
            result = result + bias
        return result



from __future__ import annotations

import torch


def round_ste(x: torch.Tensor) -> torch.Tensor:
    """Straight-through estimator for rounding."""
    return (x.round() - x).detach() + x


def clamp_ste(x: torch.Tensor, min: float | None = None, max: float | None = None) -> torch.Tensor:
    """Straight-through estimator for clamping."""
    return (x.clamp(min, max) - x).detach() + x



from __future__ import annotations

import argparse
import random

import torch
from datasets import load_dataset
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer

import paroquant.inference.backends.transformers.quantizer  # noqa: F401


_DATASETS = {
    "wikitext2": ("wikitext", {"name": "wikitext-2-raw-v1", "split": "test"}),
    "c4": (
        "allenai/c4",
        {
            "data_files": {"validation": "en/c4-validation.00000-of-00008.json.gz"},
            "split": "validation",
        },
    ),
}


def _load_tokens(dataset_name: str, seq_len: int, tokenizer, seed: int = 0) -> torch.Tensor:
    name, kwargs = _DATASETS[dataset_name]
    data = load_dataset(name, **kwargs)

    if dataset_name == "wikitext2":
        return tokenizer("\n\n".join(data["text"]), return_tensors="pt").input_ids

    rng = random.Random(seed)
    chunks = []
    for _ in range(256):
        while True:
            text = data[rng.randint(0, len(data) - 1)]["text"]
            input_ids = tokenizer(text, return_tensors="pt").input_ids
            if input_ids.shape[1] >= seq_len:
                break
        offset = rng.randint(0, input_ids.shape[1] - seq_len - 1)
        chunks.append(input_ids[:, offset : offset + seq_len])
    return torch.hstack(chunks)


@torch.no_grad()
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=str, required=True)
    parser.add_argument("--seq-len", type=int, default=2048)
    parser.add_argument("--seed", type=int, default=0)
    args = parser.parse_args()

    random.seed(args.seed)
    torch.manual_seed(args.seed)

    model = AutoModelForCausalLM.from_pretrained(args.model, torch_dtype=torch.float16, device_map="auto")
    tokenizer = AutoTokenizer.from_pretrained(args.model)
    loss_fn = torch.nn.CrossEntropyLoss().cuda()

    for dataset_name in ["wikitext2", "c4"]:
        tokens = _load_tokens(dataset_name, args.seq_len, tokenizer, args.seed)
        num_samples = tokens.numel() // args.seq_len
        tokens = tokens[0, : num_samples * args.seq_len].view(num_samples, args.seq_len)

        total_loss = 0.0
        for i in tqdm(range(num_samples), desc=dataset_name):
            input_ids = tokens[i].cuda().unsqueeze(0)
            logits = model(input_ids).logits[:, :-1, :].contiguous()
            labels = input_ids[:, 1:]
            total_loss += loss_fn(logits.view(-1, logits.size(-1)), labels.view(-1)).item()

        ppl = torch.exp(torch.tensor(total_loss / num_samples)).item()
        print(f"{dataset_name}: {ppl:.2f}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import torch
from torch import nn
from torch.utils.checkpoint import checkpoint

from .util import get_named_linears
from .quantizer import UniformAffineQuantizer

from paroquant.kernels.cuda import scaled_pairwise_rotation


class PseudoQuantizedLinear(nn.Module):
    """Pseudo-quantized linear layer using custom kernels for fine-tuning."""

    def __init__(
        self,
        linear: nn.Linear,
        rotation_pairs: list[torch.Tensor] | None,
        channel_scales: torch.Tensor | None,
        *,
        group_size: int,
        n_bits: int,
        num_rotations: int,
    ) -> None:
        super().__init__()
        self.enable_checkpoint = False
        self.weight = nn.Parameter(linear.weight.clone())
        self.in_feat = self.weight.shape[1]
        self.out_feat = self.weight.shape[0]
        num_groups = self.in_feat // group_size
        assert self.in_feat % group_size == 0
        assert self.weight.dtype == torch.float16 or self.weight.dtype == torch.float32
        if rotation_pairs is not None:
            pairs_grouped, angles_grouped, mask = rotation_pairs
            assert pairs_grouped.size(0) == num_rotations
            assert angles_grouped.size(0) == num_rotations
            assert mask.shape == angles_grouped.shape

        num_rotations = (
            torch.tensor(num_rotations, device="cuda") if not isinstance(num_rotations, torch.Tensor) else num_rotations
        )
        self.register_buffer("num_rotations", num_rotations)

        if linear.bias is None:
            self.register_buffer("bias", None)
        else:
            self.bias = nn.Parameter(linear.bias.clone())
        if rotation_pairs is None:
            # generate dummy pairs & angles if rotation_pairs is None
            angles_grouped = torch.zeros(num_rotations, self.in_feat // 2, device="cuda", dtype=self.weight.dtype)
            single_group_idx_ij = torch.randperm(group_size, device="cuda", dtype=torch.short)
            one_krot_layer_idx_ij = single_group_idx_ij.repeat(num_groups)
            pairs_grouped = one_krot_layer_idx_ij.unsqueeze(0).expand(num_rotations, -1).contiguous()
            mask = torch.zeros_like(angles_grouped, device="cuda", dtype=torch.bool)
        else:
            pairs_grouped, angles_grouped, mask = rotation_pairs
        self.pairs_grouped = nn.Parameter(pairs_grouped.clone(), requires_grad=False)
        self.angles_grouped = nn.Parameter(angles_grouped.clone())

        self.register_buffer("mask", mask)
        if channel_scales is None:
            # generate dummy channel scales
            channel_scales = torch.ones(self.in_feat, device="cuda", dtype=self.weight.dtype)
        self.channel_scales = nn.Parameter(channel_scales)
        assert self.pairs_grouped.dtype == torch.short, self.pairs_grouped.dtype
        assert self.channel_scales.dtype == self.weight.dtype, (
            self.channel_scales.dtype,
            self.weight.dtype,
        )
        assert self.mask.dtype == torch.bool, self.mask.dtype

        self.quantizer: UniformAffineQuantizer | None = None
        self.register_buffer("quantizer_optim_enabled", torch.tensor(False))
        n_bits = n_bits if isinstance(n_bits, torch.Tensor) else torch.tensor(n_bits)
        group_size = group_size if isinstance(group_size, torch.Tensor) else torch.tensor(group_size)
        self.register_buffer("n_bits", n_bits)
        self.register_buffer("group_size", group_size)

    def forward(self, x: torch.Tensor, weight_update: torch.Tensor | None = None) -> torch.Tensor:
        weight = self.weight
        if weight_update is not None:
            weight = weight + weight_update
        weight = self._pseudo_quantize(weight)
        x = self.checkpointed(torch.matmul, x, weight.T)
        if getattr(self, "bias", None) is not None:
            x = x + self.bias
        return x

    def _pseudo_quantize(self, weight: torch.Tensor) -> torch.Tensor:
        weight = weight * self.channel_scales
        weight = self.checkpointed(
            scaled_pairwise_rotation,
            weight,
            self.pairs_grouped,
            self.angles_grouped,
            None,
            self.group_size,
        )
        if self.quantizer_optim_enabled or self.quantizer is not None:
            assert self.quantizer is not None, "Quantizer should be initialized."
            self.quantizer.enable_checkpoint = self.enable_checkpoint
            weight = self.quantizer(weight)
        else:
            weight = self.checkpointed(
                UniformAffineQuantizer.pseudo_quantize,
                weight,
                self.n_bits,
                self.group_size,
            )
        if self.pairs_grouped is not None:
            flipped_pairs = torch.flip(self.pairs_grouped, dims=[0])
            flipped_angles = torch.flip(self.angles_grouped, dims=[0])
            weight = self.checkpointed(
                scaled_pairwise_rotation,
                weight,
                flipped_pairs,
                -flipped_angles,
                None,
                self.group_size,
            )
        if self.channel_scales is not None:
            weight = weight / self.channel_scales.view(1, -1)
        return weight

    def set_optim_enabled(
        self,
        *,
        weight: bool = False,
        bias: bool = False,
        angles: bool = False,
        channel_scales: bool = False,
        quantizer: bool = False,
    ) -> None:
        self.weight.requires_grad = weight
        if self.bias is not None:
            self.bias.requires_grad = bias
        if self.angles_grouped is not None:
            self.angles_grouped.requires_grad = angles
        if self.channel_scales is not None:
            self.channel_scales.requires_grad = channel_scales

        if quantizer and self.quantizer is None:
            # Initialize the quantizer with the current rotated weight
            weight = self.weight
            if self.channel_scales is not None:
                weight = weight * self.channel_scales.view(1, -1)
            if self.pairs_grouped is not None:
                weight = scaled_pairwise_rotation(
                    weight,
                    self.pairs_grouped,
                    self.angles_grouped,
                    None,
                    self.group_size,
                )

            self.quantizer = UniformAffineQuantizer(
                weight,
                n_bits=self.n_bits,
                group_size=self.group_size,
            )

        if self.quantizer:
            self.quantizer.set_optim_enabled(quantizer)
        self.quantizer_optim_enabled = torch.tensor(quantizer)

    def get_optim_params(self, name: str) -> list[nn.Parameter]:
        if name == "weight":
            return [self.weight]
        elif name == "bias":
            return [self.bias]
        elif name == "angles":
            if self.angles_grouped is None:
                return []
            return [self.angles_grouped]
        elif name == "channel_scales":
            if self.channel_scales is None:
                return []
            return [self.channel_scales]
        elif name == "quantizer":
            return self.quantizer.optim_params()
        else:
            raise ValueError(f"Unknown parameter group: {name}")

    @classmethod
    def from_state_dict(cls, state_dict: dict) -> "PseudoQuantizedLinear":
        weight: torch.Tensor = state_dict["weight"]
        has_bias: bool = "bias" in state_dict and state_dict["bias"] is not None
        linear = nn.Linear(
            weight.shape[1],
            weight.shape[0],
            bias=has_bias,
            device=weight.device,
            dtype=weight.dtype,
        )
        linear.weight.data.copy_(weight)
        if has_bias:
            linear.bias.data.copy_(state_dict["bias"])

        group_size = state_dict["group_size"]
        n_bits = state_dict["n_bits"]
        channel_scales = state_dict["channel_scales"]
        num_rotations = state_dict["num_rotations"]

        pairs_grouped = state_dict["pairs_grouped"]
        angles_grouped = state_dict["angles_grouped"]
        mask = state_dict.get("mask", torch.zeros_like(angles_grouped, dtype=torch.bool))

        qlinear = PseudoQuantizedLinear(
            linear,
            [pairs_grouped, angles_grouped, mask],
            channel_scales,
            group_size=group_size,
            n_bits=n_bits,
            num_rotations=num_rotations,
        )

        # Initialize the quantizer
        if "quantizer.scale" in state_dict:
            qlinear.set_optim_enabled(quantizer=True)

        qlinear.load_state_dict(state_dict)
        return qlinear

    @torch.no_grad()
    def reset_angles_by_mask(self):
        self.angles_grouped.data[self.mask] = 0

    def checkpointed(self, fn, *args) -> torch.Tensor:
        if torch.is_grad_enabled() and self.enable_checkpoint:
            return checkpoint(fn, *args, use_reentrant=False)
        return fn(*args)

    @torch.no_grad()
    def pseudo_weight(self) -> torch.Tensor:
        return self._pseudo_quantize(self.weight).detach()


@torch.no_grad()
def reset_angles_by_mask(module: nn.Module) -> None:
    linears = get_named_linears(module, PseudoQuantizedLinear)
    for linear in linears.values():
        linear.reset_angles_by_mask()



from __future__ import annotations

import argparse
import asyncio
import contextlib
import importlib
import io
import os
import re
import time
import warnings

from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.theme import Theme

from paroquant.inference import (
    GenerationParams,
    build_prompt,
    create_generator,
    detect_backend,
)


@contextlib.contextmanager
def _silence_stderr():
    with contextlib.redirect_stderr(io.StringIO()):
        yield


_THINKING_LINES = 4
_SPECIAL_RE = re.compile(r"<\|[^|]+\|>")


class _ThinkingTracker:
    """Parses raw streamed text, shows thinking in a small live box, then folds it."""

    _REFRESH_INTERVAL = 1.0 / 24

    def __init__(self, console: Console, enable_thinking: bool):
        self.console = console
        self.enable_thinking = enable_thinking
        self.raw = ""
        self.in_thinking = enable_thinking
        self._live: Live | None = None
        self._think_start = time.perf_counter()
        self._last_update = 0.0

    def on_token(self, text: str):
        self.raw += text

        if self.in_thinking:
            if "</think>" in self.raw:
                self.in_thinking = False
                self.stop()
                self._start_response_live()
            else:
                self._update_thinking_box()
        else:
            if self._live is None:
                self._start_response_live()
            else:
                self._update_response()

    def _update_thinking_box(self):
        now = time.perf_counter()
        if self._live is not None and now - self._last_update < self._REFRESH_INTERVAL:
            return
        self._last_update = now

        lines = self.raw.splitlines()
        tail = "\n".join(lines[-_THINKING_LINES:])
        panel = Panel(
            tail or "...",
            title=f"thinking ({now - self._think_start:.1f}s)",
            border_style="dim",
            width=min(self.console.width, 80),
            height=_THINKING_LINES + 2,
        )
        if self._live is None:
            self._live = Live(panel, console=self.console, transient=True)
            self._live.start()
        else:
            self._live.update(panel)

    def _start_response_live(self):
        self._live = Live(Markdown(""), console=self.console, vertical_overflow="visible")
        self._live.start()
        self._update_response()

    def _update_response(self):
        now = time.perf_counter()
        if now - self._last_update < self._REFRESH_INTERVAL:
            return
        self._last_update = now

        response = self._get_response()
        if response and self._live is not None:
            self._live.update(Markdown(response))

    def _get_response(self) -> str:
        if self.enable_thinking and "</think>" in self.raw:
            text = self.raw.split("</think>", 1)[1].lstrip("\n")
        elif self.in_thinking:
            text = ""
        else:
            text = self.raw
        return _SPECIAL_RE.sub("", text)

    def stop(self):
        if self._live is not None:
            response = self._get_response()
            if response:
                self._live.update(Markdown(response))
            self._live.stop()
            self._live = None

    @property
    def output_text(self) -> str:
        return self._get_response().strip()


def _suppress_library_noise():
    warnings.filterwarnings("ignore")
    os.environ.setdefault("TRANSFORMERS_NO_ADVISORY_WARNINGS", "1")
    os.environ.setdefault("HF_HUB_DISABLE_PROGRESS_BARS", "1")
    os.environ.setdefault("TOKENIZERS_PARALLELISM", "false")
    os.environ.setdefault("VLLM_LOGGING_LEVEL", "ERROR")
    for mod, fn in [
        ("huggingface_hub", "disable_progress_bars"),
        ("transformers.utils.logging", "set_verbosity_error"),
    ]:
        try:
            getattr(importlib.import_module(mod), fn)()
        except Exception:
            pass


async def run_chat_app(model: str, backend: str, params: GenerationParams):
    _suppress_library_noise()

    console = Console(theme=Theme({"hint": "dim"}))

    backend = detect_backend() if backend == "auto" else backend

    console.print(f"[hint]Loading model ({backend})...[/hint]")
    generator = create_generator(backend, model)

    console.print("[hint]Warming up...[/hint]")
    warmup_prompt = build_prompt(generator.tokenizer, [{"role": "user", "content": "Hi"}], False)
    await generator.generate(warmup_prompt, GenerationParams(max_tokens=1, temperature=0.0))

    console.clear()

    enable_thinking = False

    banner = (
        f"[bold]ParoQuant Chat[/bold]\n"
        f"Backend: [bold]{backend}[/bold]\n"
        f"Model: [bold]{model}[/bold]\n\n"
        f"[bold]/think[/bold] · [bold]/clear[/bold] · [bold]/quit[/bold]"
    )
    console.print(Panel.fit(banner, border_style="bright_blue"))

    history: list[dict[str, str]] = []

    try:
        while True:
            try:
                user_prompt = console.input(">>> ").strip()
            except (KeyboardInterrupt, EOFError):
                console.print("\n[hint]Session closed.[/hint]")
                break

            if not user_prompt:
                continue
            cmd = user_prompt.lower()
            if cmd in {"/quit", "quit", "/exit", "exit"}:
                break
            if cmd == "/clear":
                history.clear()
                console.clear()
                console.print("[hint]Conversation history cleared.[/hint]\n")
                continue
            if cmd == "/think":
                enable_thinking = not enable_thinking
                console.print(f"[hint]Thinking {'on' if enable_thinking else 'off'}.[/hint]\n")
                continue

            history.append({"role": "user", "content": user_prompt})

            prompt = build_prompt(generator.tokenizer, history, enable_thinking)

            tracker = _ThinkingTracker(console, enable_thinking)
            with _silence_stderr():
                result = await generator.generate(prompt, params, on_text=tracker.on_token)
            tracker.stop()

            history.append({"role": "assistant", "content": tracker.output_text})

            s = result.stats
            parts = []
            if s.ttft is not None:
                parts.append(f"ttft {s.ttft:.2f}s")
            parts += [f"{s.num_tokens} tokens", f"{s.tps:.1f} tok/s", f"{s.latency:.1f}s total"]
            console.print(f"  {' · '.join(parts)}", style="hint", highlight=False)
            console.print()
    finally:
        await generator.close()


def main():
    _suppress_library_noise()
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=str, required=True)
    parser.add_argument("--backend", type=str, default="auto", choices=["auto", "vllm", "transformers", "mlx"])
    parser.add_argument("--max-tokens", type=int, default=8192)
    parser.add_argument("--temperature", type=float, default=0.6)
    parser.add_argument("--top-p", type=float, default=0.95)
    parser.add_argument("--top-k", type=int, default=0)
    parser.add_argument("--repetition-penalty", type=float, default=1.0)
    args = parser.parse_args()

    asyncio.run(
        run_chat_app(
            model=args.model,
            backend=args.backend,
            params=GenerationParams(
                max_tokens=args.max_tokens,
                temperature=args.temperature,
                top_p=args.top_p,
                top_k=args.top_k,
                repetition_penalty=args.repetition_penalty,
            ),
        )
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import torch
import torch.nn as nn

# AutoAWQ's GEMM kernel imports PytorchGELUTanh from transformers.activations,
# but it was removed in transformers >=4.55. Provide a stub until AutoAWQ is updated.
import transformers.activations as _act

if not hasattr(_act, "PytorchGELUTanh"):
    _act.PytorchGELUTanh = _act.GELUActivation

from awq.modules.linear.gemm import WQLinearMMFunction


class RotateQuantizedLinear(nn.Module):
    """Pairwise Givens rotation + INT4 quantized matmul (AWQ GEMM kernel).

    All parameters are stored flat (no submodules), so state dict keys like
    ``gate_proj.theta`` and ``gate_proj.qweight`` match checkpoint naming directly.
    This enables native HuggingFace weight loading via ``from_pretrained``.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = False,
        group_size: int = 128,
        bits: int = 4,
        krot: int = 8,
    ):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.w_bit = bits
        self.group_size = group_size

        pack = 32 // bits
        n_groups = in_features // group_size

        # Rotation buffers
        self.register_buffer("theta", torch.zeros(krot, in_features // 2, dtype=torch.float16))
        self.register_buffer("pairs", torch.zeros(krot, in_features, dtype=torch.int16))
        self.register_buffer("channel_scales", torch.ones(1, in_features, dtype=torch.float16))

        # AWQ quantized weight buffers
        self.register_buffer("qweight", torch.zeros(in_features, out_features // pack, dtype=torch.int32))
        self.register_buffer("qzeros", torch.zeros(n_groups, out_features // pack, dtype=torch.int32))
        self.register_buffer("scales", torch.zeros(n_groups, out_features, dtype=torch.float16))

        if bias:
            self.register_buffer("bias", torch.zeros(out_features, dtype=torch.float16))
        else:
            self.bias = None

    @torch.no_grad()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        assert x.dtype == torch.float16, f"Expected float16 input, got {x.dtype}"
        x = torch.ops.rotation.rotate(x, self.pairs, self.theta, self.channel_scales)
        y = WQLinearMMFunction.apply(
            x,
            self.qweight,
            self.qzeros,
            self.scales,
            self.w_bit,
            self.group_size,
            self.bias,
            self.out_features,
        )
        return y.reshape(*x.shape[:-1], self.out_features)



from __future__ import annotations

from collections.abc import AsyncIterator

import mlx.core as mx
from mlx_lm.sample_utils import make_logits_processors, make_sampler

from paroquant.inference.base import BaseGenerator, GenerationParams

from .load import load


class MlxGenerator(BaseGenerator):
    backend = "mlx"

    def __init__(self, model: str):
        self.model, self.processor, self.is_vlm = load(model)
        self.tokenizer = getattr(self.processor, "tokenizer", self.processor)

    async def stream_generate(self, prompt: str, params: GenerationParams) -> AsyncIterator[str]:
        sampler = make_sampler(temp=params.temperature, top_p=params.top_p)
        logits_processors = make_logits_processors(
            repetition_penalty=params.repetition_penalty if params.repetition_penalty > 1.0 else None,
        )

        kwargs = dict(max_tokens=params.max_tokens, sampler=sampler, logits_processors=logits_processors)

        if self.is_vlm:
            from mlx_vlm import stream_generate

            if prompt:
                kwargs["input_ids"] = mx.array(self.tokenizer.encode(prompt))[None]
        else:
            from mlx_lm.generate import stream_generate

        for resp in stream_generate(self.model, self.processor, prompt, **kwargs):
            text = resp.text if hasattr(resp, "text") else resp
            if text:
                yield text



from __future__ import annotations

from paroquant.inference.base import detect_backend


def _serve_vllm():
    import asyncio
    import sys

    from vllm.entrypoints.openai.api_server import (
        FlexibleArgumentParser,
        make_arg_parser,
        run_server,
    )

    import paroquant.inference.backends.vllm.plugin  # noqa: F401

    args = make_arg_parser(FlexibleArgumentParser()).parse_args(sys.argv[1:])
    asyncio.run(run_server(args))


def _serve_mlx():
    import mlx_lm.server
    from mlx_lm.utils import load_tokenizer

    from paroquant.inference.backends.mlx.load import load as paro_load

    def _patched_load(path_or_hf_repo, tokenizer_config=None, adapter_path=None, **kwargs):
        model, _, _ = paro_load(path_or_hf_repo, force_text=True)
        tokenizer = load_tokenizer(path_or_hf_repo, tokenizer_config_extra=tokenizer_config)
        tokenizer._tool_call_start = None
        tokenizer._tool_call_end = None
        return model, tokenizer

    mlx_lm.server.load = _patched_load
    mlx_lm.server.main()


def main():
    backend = detect_backend()
    if backend in ("vllm", "transformers"):
        _serve_vllm()
    elif backend == "mlx":
        _serve_mlx()
    else:
        raise RuntimeError(f"Serve requires vllm or mlx. Detected: {backend}")


if __name__ == "__main__":
    main()



# Copyright (c) 2025, Haisheng Chen.

import torch


class RotateTensorFunc(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, idx_ij, theta, scale=None, group_size=128):
        ctx.orig_shape = x.shape
        ctx.orig_dtype = x.dtype
        ctx.has_scale = scale is not None
        ctx.group_size = group_size

        y = torch.ops.rotation.rotate(x, idx_ij, theta, scale, group_size)
        saved = (x, idx_ij, theta, y, scale) if ctx.has_scale else (x, idx_ij, theta, y)
        ctx.save_for_backward(*saved)
        return y

    @staticmethod
    def backward(ctx, grad_out):
        x, idx_ij, theta, y = ctx.saved_tensors[:4]
        scale = ctx.saved_tensors[4] if ctx.has_scale else None
        gs = ctx.group_size

        KROT, H = idx_ij.shape
        num_groups = H // gs
        half_gs = gs // 2
        B = y.numel() // H
        t = y.view(B, H)
        g = grad_out.view(-1, H)
        grad_theta = torch.zeros_like(theta)
        offsets = torch.arange(num_groups, device=idx_ij.device, dtype=torch.int32).unsqueeze(1) * gs

        for i in range(KROT - 1, -1, -1):
            neg_theta = -theta[[i]]
            idx = idx_ij[[i]]
            t = torch.ops.rotation.rotate(t, idx, neg_theta, None, gs)
            g = torch.ops.rotation.rotate(g, idx, neg_theta, None, gs)

            idx = idx.view(num_groups, gs)
            di = (idx[:, 0::2] + offsets).reshape(-1)
            dj = (idx[:, 1::2] + offsets).reshape(-1)

            a = t[:, di].view(B, num_groups, half_gs)
            b = t[:, dj].view(B, num_groups, half_gs)
            ga = g[:, di].view(B, num_groups, half_gs)
            gb = g[:, dj].view(B, num_groups, half_gs)

            sin_t, cos_t = theta[i].view(num_groups, half_gs).sin(), theta[i].view(num_groups, half_gs).cos()
            grad_theta[i] = (
                ((ga * b - gb * a).sum(0) * cos_t - (ga * a + gb * b).sum(0) * sin_t).reshape(-1).to(theta.dtype)
            )

        if ctx.has_scale:
            grad_x = (g * scale.unsqueeze(0)).view(ctx.orig_shape).to(ctx.orig_dtype)
            grad_scale = (x.view(B, H) * g).sum(0).to(dtype=scale.dtype, device=scale.device)
        else:
            grad_x = g.view(ctx.orig_shape).to(ctx.orig_dtype)
            grad_scale = None

        return grad_x, None, grad_theta, grad_scale, None


def scaled_pairwise_rotation(x, idx_ij, theta, scales=None, group_size=128):
    return RotateTensorFunc.apply(x, idx_ij, theta, scales, group_size)



from __future__ import annotations

import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal

import simple_parsing
import torch
import torch.nn as nn
from tqdm import tqdm

from paroquant.optim.train import (
    optimize_module,
    get_random_rotation_pairs,
)
from paroquant.optim.qlinear import (
    PseudoQuantizedLinear,
    reset_angles_by_mask,
)
from paroquant.optim.util import (
    set_module_by_name,
    load_model,
    move_embed,
    load_tokenizer,
    get_blocks,
    get_calib_dataset,
    get_mixed_calib_dataset,
    catch_first_layer_input,
    get_named_linears,
    empty_cache,
    logger,
    CachedTensorShards,
)
from paroquant.optim.rotation import transform_to_kernel_data


@dataclass(kw_only=True)
class Config:
    # Huggingface model path.
    model: str
    # The parameters to optimize at each stage and the corresponding learning rates,
    # e.g., --params "channel_scales:0.05,angles:0.05" "weight:1e-5,quantizer:1e-6"
    params: list[str]
    # The number of epochs for each stage of optimization,
    # e.g., --epochs 10 10
    epochs: list[int]

    weight_decay: float = 0.01
    betas: tuple[float, float] = (0.9, 0.95)
    eps: float = 1e-10
    # Loss function to use.
    loss: Literal["mse", "smooth_l1"] = "smooth_l1"

    # Quantization & rotation group size.
    group_size: int
    # Bit width.
    n_bit: int
    # Number of rotations.
    num_rotations: int

    skipped_modules: list[str] = field(default_factory=list)

    # Calibration datasets. If more than one dataset is provided,
    # they will be sampled evenly and shuffled.
    datasets: list[str]
    val_dataset: str
    train_size: int
    validation_size: int
    batch_size: int
    val_batch_size: int | None = None
    seqlen: int

    # Number of shards to cache the input/output tensors. At any time, only one shard
    # will be moved to GPU for optimization. The rest will be kept in CPU memory.
    # Increasing this reduces GPU memory usage but increases training time.
    cache_shards: int = 1

    # Directory to save state dicts of optimized linear layers.
    output_dir: str

    # Whether to resume from previously saved results in `output_dir`.
    resume: bool = False
    # Whether to enable gradient checkpointing.
    checkpointing: bool = False

    seed: int


def main():
    args = simple_parsing.parse(Config, add_option_string_dash_variants=simple_parsing.DashVariant.DASH)
    print(args)

    # Store the results in a subdirectory.
    model_name = args.model.split("/")[-1]
    output_dir = Path(args.output_dir)
    output_dir = output_dir / model_name
    output_dir.mkdir(parents=True, exist_ok=True)

    # Currently only support single GPU training.
    device = "cuda"

    # Determine which params to optimize.
    params_to_optimize: list[dict[str, float]] = []
    for params in args.params:
        params = params.strip().split(",")
        param_dict = {}
        for param in params:
            param, lr = param.strip().split(":")
            param_dict[param.strip()] = float(lr.strip())
        params_to_optimize.append(param_dict)
    print(f"Parameters to optimize: {params_to_optimize}")

    # Save args to output directory
    with open(output_dir / "args.json", "w") as f:
        json.dump(vars(args), f, indent=2)

    # Load model.
    model = load_model(args.model, device_map="cpu", dtype=torch.float16)
    move_embed(model, device)
    tokenizer = load_tokenizer(args.model)
    blocks = get_blocks(model)

    # Get calibration dataset.
    samples = get_mixed_calib_dataset(
        args.datasets,
        tokenizer=tokenizer,
        n_samples=args.train_size,
        block_size=args.seqlen,
        seed=args.seed,
        split="train",
    )
    samples = torch.stack(samples, dim=0).to(device)

    val_samples = get_calib_dataset(
        args.val_dataset,
        tokenizer=tokenizer,
        n_samples=args.validation_size,
        block_size=args.seqlen,
        seed=args.seed,
        split="validation",
    )
    val_samples = torch.stack(val_samples, dim=0).to(device)

    # Capture first layer's input.
    logger.info("Capturing first layer input...")
    blocks[0].to(device)
    og_layer_input_batches, kwargs = catch_first_layer_input(
        model,
        blocks,
        samples,
        batch_size=args.batch_size,
    )
    val_batch_size = args.val_batch_size or args.batch_size
    og_layer_val_input_batches, _ = catch_first_layer_input(
        model,
        blocks,
        val_samples,
        batch_size=val_batch_size,
    )
    blocks[0].cpu()

    del samples
    empty_cache()

    @torch.no_grad()
    def forward_layer_batch(
        layer: nn.Module,
        input_batched: list[torch.Tensor],
        kwargs: dict,
        store_device: torch.device,
        dtype: torch.dtype = torch.float16,
        cast_to_dtype: torch.dtype = torch.float16,
    ) -> list[torch.Tensor]:
        output_batched = []

        layer.to(device)
        for input_batch in input_batched:
            output = layer(input_batch.to(dtype).to(device), **kwargs)
            if isinstance(output, tuple):
                output = output[0]
            if output.dtype != cast_to_dtype:
                output = output.to(cast_to_dtype)
            if output.device != store_device:
                output = output.to(store_device)
            output_batched.append(output)
        layer.cpu()

        empty_cache()
        return output_batched

    def set_checkpointing_enabled(module: nn.Module, enable: bool) -> None:
        for linear in get_named_linears(module, subclass=PseudoQuantizedLinear).values():
            linear.enable_checkpoint = enable

    # Layerwise, multi-stage optimization.
    for layer_idx, layer in enumerate(tqdm(blocks)):
        empty_cache()
        logger.info(f"Capturing original layer output...")
        # Original output of this layer.
        og_layer_output_batches = forward_layer_batch(layer, og_layer_input_batches, kwargs, store_device="cpu")
        og_layer_val_output_batches = forward_layer_batch(layer, og_layer_val_input_batches, kwargs, store_device="cpu")

        if layer_idx > 0:
            layer_input_batches = new_layer_output_batches
            layer_val_input_batches = new_layer_val_output_batches
        else:
            layer_input_batches = og_layer_input_batches
            layer_val_input_batches = og_layer_val_input_batches

        train_input_batches = layer_input_batches
        train_output_batches = og_layer_output_batches

        train_input_batches = CachedTensorShards(train_input_batches, args.cache_shards, target_device=device)
        train_output_batches = CachedTensorShards(train_output_batches, args.cache_shards, target_device=device)

        val_input_batches = [b.to(device) for b in layer_val_input_batches]
        val_output_batches = [b.to(device) for b in og_layer_val_output_batches]

        # Freeze all parameters
        for param in layer.parameters():
            param.requires_grad = False

        linear_modules = get_named_linears(layer)
        if args.resume:
            all_files_exist = True
            for name in linear_modules.keys():
                file_name = f"{layer_idx}.{name}.pt"
                file_path = output_dir / file_name
                if not file_path.exists() and name not in args.skipped_modules:
                    all_files_exist = False
                    break
        else:
            all_files_exist = False

        if not all_files_exist:
            logger.info(f"Initializing rotation parameters...")

        for name, old_module in linear_modules.items():
            if name in args.skipped_modules:
                continue

            if all_files_exist:
                existing_result_file = output_dir / f"{layer_idx}.{name}.pt"
                sd = torch.load(existing_result_file, map_location=device)
                new_module = PseudoQuantizedLinear.from_state_dict(sd)
                set_module_by_name(layer, name, new_module)
                continue

            old_module.to(device)

            num_pairs_factor = 0.5
            weight = old_module.weight.float()
            weight_grouped = weight.view(weight.shape[0], -1, args.group_size).permute(1, 0, 2)

            all_pairs = get_random_rotation_pairs(
                weight_grouped,
                group_size=args.group_size,
                num_rotations=args.num_rotations,
                num_pairs_factor=num_pairs_factor,
                seed=args.seed + layer_idx,
            )

            all_pairs = [torch.tensor(pairs, device="cpu", dtype=torch.int32) for pairs in all_pairs]
            initial_angles = [torch.zeros(pairs.shape[0], device="cpu") for pairs in all_pairs]
            initial_scales = torch.ones(1, weight.shape[1], dtype=torch.float16, device=device)

            npairs, angles, mask = transform_to_kernel_data(
                all_pairs,
                initial_angles,
                group_size=args.group_size,
            )
            npairs = npairs.to(device)
            angles = angles.to(device)
            mask = mask.to(device)
            rotation_pairs = [npairs, angles, mask]
            channel_scales = initial_scales

            new_module = PseudoQuantizedLinear(
                old_module,
                rotation_pairs,
                channel_scales,
                group_size=args.group_size,
                n_bits=args.n_bit,
                num_rotations=args.num_rotations,
            )

            set_module_by_name(layer, name, new_module)
            old_module.cpu()

        if not all_files_exist:
            layer.to(device).float()

            set_checkpointing_enabled(layer, args.checkpointing)
            for step, step_params_dict in enumerate(params_to_optimize):
                empty_cache()
                optim_params = []
                new_modules = get_named_linears(layer, subclass=PseudoQuantizedLinear)
                for new_module in new_modules.values():
                    new_module.set_optim_enabled(
                        **{param_name: True for param_name in step_params_dict.keys()},
                    )
                    for param_name, lr in step_params_dict.items():
                        optim_params.append(
                            dict(
                                params=new_module.get_optim_params(param_name),
                                lr=lr,
                                weight_decay=args.weight_decay,
                                betas=args.betas,
                                eps=args.eps,
                            )
                        )

                logger.info(
                    f"Optimizing layer {layer_idx}, step {step + 1}/{len(params_to_optimize)}: "
                    f"{', '.join([k for k in step_params_dict])}"
                )

                optimize_module(
                    layer,
                    (train_input_batches, train_output_batches),
                    (val_input_batches, val_output_batches),
                    kwargs,
                    optim_params,
                    loss_fn=args.loss,
                    n_iter=args.epochs[step],
                    early_stop=None,
                    post_optim_callback=reset_angles_by_mask,
                )

            set_checkpointing_enabled(layer, False)

            del (
                train_input_batches,
                train_output_batches,
                val_input_batches,
                val_output_batches,
            )
            empty_cache()

        else:
            logger.info(f"Skipping optimization for layer {layer_idx}: already been optimized.")

        layer.half().to(device)

        logger.info("Capturing new layer output...")
        new_layer_output_batches = forward_layer_batch(
            layer,
            layer_input_batches,
            kwargs,
            store_device="cpu",
        )
        new_layer_val_output_batches = forward_layer_batch(
            layer,
            layer_val_input_batches,
            kwargs,
            store_device="cpu",
        )

        og_layer_input_batches = og_layer_output_batches
        og_layer_val_input_batches = og_layer_val_output_batches

        if all_files_exist:
            layer.cpu()
            continue

        # Save the optimized result
        for name, module in get_named_linears(layer, subclass=PseudoQuantizedLinear).items():
            result_file = output_dir / f"{layer_idx}.{name}.pt"
            torch.save(
                module.state_dict(),
                result_file,
            )

        layer.cpu()


if __name__ == "__main__":
    main()



from .base import build_prompt, create_generator, detect_backend, GenerationParams, GenerationResult, GenerationStats

__all__ = [
    "build_prompt",
    "create_generator",
    "detect_backend",
    "GenerationParams",
    "GenerationResult",
    "GenerationStats",
]
